#ifndef LIBLINUX_H
#define LIBLINUX_H


// Some replacement routines missing in gcc

#define _NORMALCURSOR 1
#define _NOCURSOR 0
#define O_BINARY O_RDWR

#ifdef USE_POGO
#include <stdlib.h>
#include <string.h>

#include <pogo/pogo.h>

void rewind(POGOFILE *stream);
int set_window(int);

enum {PLAY_WIN, STAT_WIN, MAP_WIN, TEXT_WIN, MSG_WIN};

void init_libpogo();

enum {ACT_NONE, ACT_NEW_LEVEL, ACT_SHOUT, ACT_ATTACK, ACT_MAGIC, ACT_HEAL, ACT_DEATH, ACT_HISCORE};

int action_hook(int a, int a0=-1, int a1=-1, int a2=-1);

void mark_square(int x, int y);

void lock_shortcuts(char *keys);
void set_direction_mode(int n, int lockt = -1);
void set_menu_mode(int n, int miny = -1, int maxy = -1, int x2 = -1);

POGOFILE *pogo_smartopen(const char *name, const char *flg);

#define TEXT_WIDTH 51
#define TEXT_HEIGHT 24
#define TILES_WIDTH 21
#define TILES_HEIGHT 16

#define BEAMX (-7)

enum { PT_SWORD = 128, PT_ARMOUR, PT_FISH, PT_SCROLL, PT_POTION, PT_DOOR, PT_OPEN_DOOR, PT_STAIR_DOWN,
	PT_STAIR_UP, PT_TRAP, PT_BOOK, PT_STONE_WALL, PT_FLOOR, PT_WAND, PT_RING, PT_AMULET, PT_HUMAN,
	PT_WHIP, PT_ARROW, PT_BOW, PT_CLUB, PT_KNIFE, PT_STONES, PT_SHIELD, PT_AXE, PT_STAFF, PT_BRICK_WALL,
	PT_BRICK_WALL2, PT_BLOCK_WALL, PT_WATER, PT_WATER2, PT_ENTRANCE, PT_HOUSE, PT_STATUE, PT_ALTAR,
	PT_FOUNTAIN, PT_GEM, PT_CORPSE, PT_STONE_WALL2, PT_STONE };


#if (defined CYGWIN) || (defined DS)

inline int random()
{
    return rand();
}

inline void srandom(unsigned int seed)
{
    srand(seed);
}


#else

#define stricmp strcasecmp

#endif

void pogo_perror(const char *str);

//int pogo_time(void *p);

#endif


char getche(void);


int clrscr(void);
int cprintf(const char *format,...);
int gotoxy(int x, int y);
int kbhit(void);
int key_to_command(int);
int putch(unsigned char chr);

#ifndef CYGWIN
int stricmp(const char *str1, const char *str2);
//int itoa(int value, char *strptr, int radix);
char *strlwr(char *str);
int itoa(int value, char *strptr, int radix);
#endif


int translate_keypad(int keyin);
int wherex(void);
int wherey(void);
int window(int x1, int y1, int x2, int y2);
void update_screen(void);
void clear_to_end_of_line(void);
void clear_to_end_of_screen(void);
int get_number_of_lines_from_curses(void);
void get_input_line_from_curses( char *const buff, int len );

void _setcursortype(int curstype);
void delay(unsigned long time);
void init_key_to_command();
void lincurses_shutdown(void);
void lincurses_startup(void);
void textbackground(int bg);
void textcolor(int col);



#ifndef _LIBLINUX_IMPLEMENTATION
/* Some stuff from curses, to remove compiling warnings.. */
extern "C"
{
    int getstr(char *);
    int pogo_getch(void);
    int noecho(void);
    int echo(void);
}
#endif


#endif
