
#include "chism/gba_nds_fat.h"
#undef PACKED
#define NO_POGO_TYPES
typedef unsigned char uchar;
#include <pogo/device.h>

int dprintf(char *fmt, ...);

//#define FAT_chdir FAT_CWD

PogoDevice fatdev;

static FAT_FILE *open_files[32];
static int open_count = 0;

char tmpname[256];

int fatdev_open(const char *name, int flags)
{
	FAT_chdir("/");

	while(*name == '/') name++;
	sprintf(tmpname, "/%s", name);

	FAT_FILE *f = FAT_fopen(tmpname, "r");
	if(f)
	{
		open_files[open_count++] = f;
		return open_count - 1;
	}
	return -1;
}

int fatdev_read(int fd, void *dest, int size)
{
	FAT_FILE *f = open_files[fd];
	dprintf("fatread %d", size);
	return FAT_fread(dest, 1, size, f);
}

int fatdev_write(int fd, const void *dest, int size)
{
	FAT_FILE *f = open_files[fd];
	return FAT_fwrite(dest, 1, size, f);
}

void fatdev_close(int fd)
{
	FAT_FILE *f = open_files[fd];
	FAT_fclose(f);
}

int fatdev_lseek(int fd, int offset, int origin)
{
	//char *ONAMES[3] = { "SET", "CUR", "END" };

	FAT_FILE *f = open_files[fd];
//	dprintf("Fatseek %d %s", offset, ONAMES[origin]);

	FAT_fseek(f, offset, origin);
	return FAT_ftell(f);
}

int fatdev_tell(int fd)
{
	FAT_FILE *f = open_files[fd];
	return FAT_ftell(f);
}

int fatdev_ioctl(int fd, int request, ...)
{
	return -1;
}

int fatdev_remove(const char *name)
{
	return -1;
}

static int count = 0;

int fatdev_dopen(const char *name)
{
	while(*name == '/') name++;
	sprintf(tmpname, "/%s", name);
	FAT_chdir(tmpname);
	count = 0;
	return 0;
}

int fatdev_dread(int fd, void *d, int len)
{
	char dest[256];
	char *ptr;
	FILE_TYPE rc;

	if(!count)
		rc = FAT_FindFirstFileLFN(dest);
	else
		rc = FAT_FindNextFileLFN(dest);

	ptr = dest;

	if(!*dest)
	{
		strcpy(dest, "(null)");
	}
	else
	{
		while(*ptr)
		{
			if(*ptr < 0x20 || *ptr > 0x7e)
				*ptr = 'X';
			ptr++;
		}
	}

	count++;

	strcpy((char *)d, dest);

	if(rc == FT_NONE)
		return 0;
	//else
	//	FAT_GetLongFilename(dest);

	return strlen(dest);
}

void fatdev_dclose(int fd)
{
}

int fatdev_stat(const char *name, struct pogo_stat *buffer)
{
	u32 len;
	FAT_chdir("/");

	while(*name == '/') name++;
	sprintf(tmpname, "/%s", name);

	if(FAT_FileExists(tmpname))
	{
		buffer->st_size = FAT_GetFileSize();
		buffer->st_mode = 0;
		if(FAT_GetFileAttributes() & ATTRIB_DIR)
			buffer->st_mode = PGS_IFDIR;
		return 0;
	}
	return -1;
}

void fatdev_init()
{
	memset(&fatdev, 0, sizeof(PogoDevice));
	fatdev.open = fatdev_open;
	fatdev.read = fatdev_read;
	fatdev.write = fatdev_write;
	fatdev.close = fatdev_close;
	fatdev.lseek = fatdev_lseek;
	fatdev.stat = fatdev_stat;
	fatdev.dopen = fatdev_dopen;
	fatdev.dread = fatdev_dread;
	fatdev.dclose = fatdev_dclose;

	FAT_InitFiles();

	pogo_device_register(&fatdev, "/fat", NULL, -1);
}
