#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <pogo/types.h>
#include <pogo/device.h>
#include <pogo/readline.h>
#include <pogo/io.h>

//#include <sys/types.h>
#include <sys/stat.h>

//#include <dirent.h>
//#include <unistd.h>

#define UNIMP(x) return x;

#include <ctype.h>


#include <pogo/font.h>
#include <pogo/console.h>

#include "glk.h"
#include "gi_blorb.h"  
#include "gi_dispa.h"

#include "dict.h"

#include <assert.h>

void console_init();
void console_set_target(void *pixels, int stride);
int console_write(int fd, const void *dest, int size);
int console_window_create(void *pixels, int w, int h, int stride, Font **fonts, int type);
int console_ioctl(int fd, int req, ...);

void readline_start();
char *readline_end(int);
int readline_putchar(int c);
void readline_set(char *str);

#define TRUE 1
#define FALSE 0

enum { FONT_FIXED, FONT_NORMAL, FONT_ITALIC, FONT_BOLD };

struct glk_window_struct
{
	uint16 x;
	uint16 y;
	uint16 width;
	uint16 height;
	int console_win;
	winid_t kids[2]; // Negative if leaf window, kids if parent
	glui32 rock;
	winid_t parent;
	
	strid_t stream;
	strid_t echo;

	int wintype;
	int method;
	int size;

	winid_t keywin;

	gidispatch_rock_t disprock;

};


gidispatch_rock_t (*register_dispatch)(void *obj, glui32 objclass) = NULL;
void (*unregister_dispatch)(void *obj, glui32 objclass, gidispatch_rock_t objrock) = NULL;

winid_t rootwin = NULL;

Font *type_fonts[4] = { NULL, NULL, NULL, NULL };
static int font_width;
static int font_height;

static int listen_flags = 0;
#define CHAR_EVENT 1
#define LINE_EVENT 2
#define MOUSE_EVENT 4

static char *linebuf = NULL;
static winid_t linewin;

typedef struct
{
	glui32 fg[style_NUMSTYLES];
	glui32 bg[style_NUMSTYLES];
} WinStyle;

WinStyle winstyles[6];

#define setwin(win) console_ioctl(1, CC_SETWIN, win->console_win)

struct glk_stream_struct
{
	winid_t window;		// If set this is a window stream
	FILE *fp;			// File stream
	glui32 rock;
	uint32 flags;
	unsigned char *mem;
	int mempos;
	int memlen;

	gidispatch_rock_t disprock;
};

strid_t active_stream = NULL;

strid_t new_stream(int rock)
{
	strid_t str = (strid_t)malloc(sizeof(struct glk_stream_struct));
	str->window = NULL;
	str->fp = NULL;
	str->rock = rock;
	str->flags = 0;
	str->mem = NULL;
	str->mempos = 0;
	str->memlen = 0;

    if (register_dispatch)
        str->disprock = (*register_dispatch)(str, gidisp_Class_Stream);
	else
		str->disprock.ptr = NULL;


	return str;
}

void free_stream(strid_t str)
{
    if (unregister_dispatch)
	{
        (*unregister_dispatch)(str, gidisp_Class_Stream, str->disprock);
        str->disprock.ptr = NULL;
    }

	free(str);
}

struct glk_fileref_struct
{
	char *name;
	int mode;
	glui32 rock;
	FILE *fp;

	gidispatch_rock_t disprock;
};

frefid_t new_fileref(int rock)
{
	frefid_t f = (frefid_t)malloc(sizeof(struct glk_fileref_struct));
	f->rock = rock;
	f->fp = NULL;
	f->name = NULL;
	f->mode = 0;

    if (register_dispatch)
        f->disprock = (*register_dispatch)(f, gidisp_Class_Fileref);
	else
		f->disprock.ptr = NULL;

	return f;
}




int screen_width;
int screen_height;
void *screen_pixels;
int screen_stride;

int graphics_init(void **pixels, int *width, int *height, int *stride);
void graphics_flush();
void graphics_exit();
int events_get(event_t *glkevt);
int events_getmsec();

void io_init();

static void window_resize(winid_t root);


void rearrange()
{
	graphics_init(&screen_pixels, &screen_width, &screen_height, &screen_stride);
	console_set_target(screen_pixels, screen_stride);
	if(rootwin)
	{
		rootwin->width = screen_width;
		rootwin->height = screen_height;
		window_resize(rootwin);
	}
	graphics_flush();
}


static void handle_more()
{
	event_t e;
	//console_write(1, "[MORE]", 6);
	//console_ioctl(1, CC_FLUSH);
	graphics_flush();
	e.type = evtype_None;
	while(e.type != evtype_CharInput)
	{
		events_get(&e);
		if(e.type == evtype_Arrange)
			rearrange();
	}
	//console_ioctl(1, CC_ERASE, 6);
}



void glk_exit(void)
{
	graphics_flush();
	graphics_exit();
	exit(0);
}

char *find_file(const char *name, const char *path);


int glk_init(int argc, char **argv)
{
	int i;
	char *fname;

	graphics_init(&screen_pixels, &screen_width, &screen_height, &screen_stride);
	io_init();

	console_set_target(screen_pixels, screen_stride);
	console_ioctl(1, CC_SETMORECALLBACK, handle_more);

	for(i=0; i<6; i++)
	{
		glui32 fg, bg;
		int j;

		if(i != wintype_TextGrid)
		{
			fg = 0x000000;
			bg = 0xcccccc;
		}
		else
		{
			bg = 0x555555;
			fg = 0xffffff;
		}

		for(j=0; j<style_NUMSTYLES; j++)
		{
			winstyles[i].fg[j] = fg;
			winstyles[i].bg[j] = bg;
		}
	}


	if((fname = find_file("Fixed.font", POGO_NULL)))
		type_fonts[0] = font_load(fname);
	if((fname = find_file("Normal.font", POGO_NULL)))
		type_fonts[1] = font_load(fname);
	if((fname = find_file("Italic.font", POGO_NULL)))
		type_fonts[2] = font_load(fname);
	if((fname = find_file("Bold.font", POGO_NULL)))
		type_fonts[3] = font_load(fname);

	if(type_fonts[0])
		type_fonts[0]->flags &= ~FFLG_TRANSP;

	for(i=0; i<4; i++)
		if(!type_fonts[i])
			type_fonts[i] = font_getdefault();
	
	
	readline_set_dict((const char **)verbs, sizeof(verbs)/sizeof(char *));
	
	return 0;

}

int do_redraw = 0;

#define gotoxy(x, y) console_ioctl(1, CC_GOTOXY, x, y)
#define cprint(s) { char *t = s; console_write(1, t, strlen(t)); }
#define setcolor(f, b) console_ioctl(1, CC_SETCOLOR, f, b)

typedef struct
{
	char *name;
	int size;

} FileInfo;

FileInfo *flist;

static POGODIR *dir_d;
static struct pogo_dirent *dir_de;
//static char dirbuf_d[40];

void glk_opendir(char *dir)
{
	dir_d = NULL;
	if(dir && strlen(dir))
		dir_d = pogo_opendir(dir);
}

char *glk_readdir()
{
	if(dir_d)
	{
		if((dir_de = pogo_readdir(dir_d)))
			return dir_de->d_name;
	}
	return NULL;
}

void glk_closedir()
{
	if(dir_d)
		pogo_closedir(dir_d);
}


#ifdef _WIN32
#define FILE_SEP '\\'
#else
#define FILE_SEP '/'
#endif

int filesort_func(const void *a, const void *b)
{
	FileInfo *f0 = (FileInfo*)a;
	FileInfo *f1 = (FileInfo*)b;
	
	if(f0->size >= 0 && f1->size < 0) return 1;
	if(f0->size < 0 && f1->size >= 0) return -1;		
	return strcmp(f0->name, f1->name);
}

int file_request(char *titlename, char *dir, int dirlen, char *filename, int filelen)
{
	char tmp[128];
	int i,pos,start,fh,columns;
	int mainw,entry,title;
	int marked = 0;
	//int fg = 0;
	//int bg = 15;
	int count;
	char *dirfile;
	struct pogo_stat ss;
	int newdir;
	int lines;
	int done = 0;
	Font *fonts[4] = {0,0,0,0};

	flist = malloc(sizeof(FileInfo) * 256);
	memset(flist, 0, sizeof(FileInfo) * 256);

	fonts[0] = font_getdefault();

	// Just to clear the entire background
	mainw = console_ioctl(1, CC_OPENWIN, 0, 0, screen_width, screen_height, fonts, CONWIN_FIXED);
	setcolor(CCOL_BLACK, CCOL_BLACK);
	console_ioctl(1, CC_CLEAR);
	console_ioctl(1, CC_CLOSEWIN, mainw);

	fh = fonts[0]->height;
	columns = screen_width / (fonts[0]->charwidth + fonts[0]->spacing);
	lines = (screen_height-6)/fh - 2;

	title = console_ioctl(1, CC_OPENWIN, 2, 2, screen_width-4, fh, fonts, CONWIN_FIXED);
	setcolor(CCOL_WHITE, CCOL_RED);
	mainw = console_ioctl(1, CC_OPENWIN, 2, fh+4, screen_width-4, fh*lines, fonts, CONWIN_FIXED);
	setcolor(CCOL_BLACK, CCOL_GRAY);
	entry = console_ioctl(1, CC_OPENWIN, 2, fh+fh*lines+6, screen_width-4, fh, fonts, CONWIN_FIXED);
	setcolor(CCOL_BLACK, CCOL_GRAY);
	console_ioctl(1, CC_CLEAR);
	readline_start(200);
	readline_set(filename);

	//lines = (screen_height/fh)-6;

	while(!done)
	{
		count = 0;
		glk_opendir(dir);
		while((dirfile = glk_readdir()))
		{
			if(strcmp(dirfile, ".") != 0)
			{
				int rc,slen;
				sprintf(tmp, "%s%s", dir, dirfile);
				if(dir && strlen(dir))
					rc = pogo_stat(tmp, &ss);
				
				slen = strlen(dirfile);
				
				if(flist[count].name)
				{
					if(strlen(flist[count].name) < slen)
						flist[count].name = realloc(flist[count].name, slen + 1);
						
				}
				else
					flist[count].name = malloc(slen + 1);
				strcpy(flist[count].name, dirfile);
				if(rc < 0)
					flist[count].size = 0;
				else
				{
					if(ss.pst_mode & PGS_IFDIR)
						flist[count].size = -1;
					else
						flist[count].size = ss.pst_size;
				}
				count++;
			}
		}
		//closedir(d);
		glk_closedir();
		
		qsort(flist, count, sizeof(FileInfo), filesort_func);

		pos = 0;
		start = 0;
		newdir = 0;
		marked = 0;

		while(!newdir && !done)
		{
			event_t evt;
			int size = lines;
			if(count < size)
				size = count;

			console_ioctl(1, CC_SETWIN, title);
			console_ioctl(1, CC_CLEAR);
			sprintf(tmp, " %s (%s)", titlename, dir);
			cprint(tmp);

			console_ioctl(1, CC_SETWIN, mainw);
			console_ioctl(1, CC_CLEAR);


			for(i=0; i<size; i++)
			{			
				pos = i + start;
				if(marked == pos)
					setcolor(CCOL_BLUE, CCOL_GRAY);
	
				console_ioctl(1, CC_GOTOXY, 0, i);

				if(flist[pos].size == -1)
					sprintf(tmp, "%c [%s]", (marked == pos) ? '>' : ' ', flist[pos].name);
				else
					sprintf(tmp, "%c  %s", (marked == pos) ? '>' : ' ', flist[pos].name);
				cprint(tmp);
				
				if(flist[pos].size != -1)
				{
					sprintf(tmp, "%10d KB", flist[pos].size/1024);
					console_ioctl(1, CC_GOTOXY, columns-14, i);
					cprint(tmp);
				}

				if(marked == pos)
					setcolor(CCOL_BLACK, CCOL_GRAY);

				//if(i<size-1)
				//	cprint("\n");
			}

			graphics_flush();

			evt.type = evtype_None;
			while(evt.type != evtype_CharInput)
			{
				int rc = events_get(&evt);
				if(evt.type == evtype_Arrange)
					rearrange();
				if(!rc)
					evt.type = evtype_None;
			}

			if(evt.val1 == keycode_Up)
				marked--;
			else
			if(evt.val1 == keycode_Down)
				marked++;
			else
			if(evt.val1 == keycode_PageUp)
				marked-=lines;
			else
			if(evt.val1 == keycode_PageDown)
				marked+=lines;
			else
			if(evt.val1 == keycode_Escape)
				done = 1;
			else
			if(evt.val1 == ' ')
			{
				if(flist[marked].size == -1)
				{
					if(strcmp(flist[marked].name, "..") == 0)
					{
						char *endp = &dir[strlen(dir)-2];
						while(*endp && *endp != FILE_SEP)
						{
							*endp = 0;
							endp--;
						}
					}
					else
					{
						char *ptr;
						strcat(dir, flist[marked].name);
						ptr = &dir[strlen(dir)];
						ptr[0] = FILE_SEP;
						ptr[1] = 0;
					}
					newdir = 1;
				}
				else
				{
					console_ioctl(1, CC_SETWIN, entry);
					readline_set(flist[marked].name);
					console_ioctl(1, CC_SETWIN, mainw);
				}
			}
			else
			if(evt.val1 == keycode_Return)
			{
				done = 2;
			}
			else
			{
				int v = evt.val1;
				switch(v)
				{
				case keycode_Left: v = RL_KEY_LEFT; break;
				case keycode_Right: v = RL_KEY_RIGHT; break;
				case keycode_Delete: v = RL_KEY_DEL; break;
				default:
					if(v < 0x20 || v > 0x7f)
						v = -1;
				}

				console_ioctl(1, CC_SETWIN, entry);
				if((v>=0) && readline_putchar(v))
					graphics_flush();
				console_ioctl(1, CC_SETWIN, mainw);
			}


			if(marked < 0)
				marked = 0;
			if(marked >= count)
				marked = count-1;

			while(marked < start)
				start--;
			while(marked >= start + size)
				start++;

		}
	}

	strcpy(filename, readline_end(0));

	if(active_stream && active_stream->window)
		setwin(active_stream->window);

	console_ioctl(1, CC_CLOSEWIN, mainw);
	console_ioctl(1, CC_CLOSEWIN, entry);
	console_ioctl(1, CC_CLOSEWIN, title);

	do_redraw = 1;
	
	i = 0;
	while(flist[i].name)
		free(flist[i++].name);
	
	free(flist);

	return done-1;

}

void glk_set_interrupt_handler(void (*func)(void))
{
	//
}

void glk_tick(void)
{
	//
}

glui32 glk_gestalt(glui32 sel, glui32 val)
{
	return glk_gestalt_ext(sel, val, NULL, 0);
}

glui32 glk_gestalt_ext(glui32 sel, glui32 val, glui32 *arr, glui32 arrlen)
{
	switch(sel)
	{
	case gestalt_Version:
		return 0x00000601;

	case gestalt_LineInput:
		if(val >= 32 && val < 127)
			return TRUE;
		else
			return FALSE;
                
	case gestalt_CharInput: 
		if(val >= 32 && val < 127)
			return TRUE;
		else if(val == keycode_Return)
			return TRUE;
		else
			return FALSE;
        
	case gestalt_CharOutput: 
		if(val >= 32 && val < 127)
		{
			if(arr && arrlen >= 1)
				arr[0] = 1;
			return gestalt_CharOutput_ExactPrint;
		}
		else
		{
			if(arr && arrlen >= 1)
				arr[0] = 1;
			return gestalt_CharOutput_CannotPrint;
		}
	case gestalt_Graphics:
		return TRUE;
	case gestalt_Timer:
		return TRUE;
	case gestalt_DrawImage:
		if(val == wintype_Graphics)
			return TRUE;
		else
			return FALSE;
	}
	
	return FALSE;

}

unsigned char glk_char_to_lower(unsigned char ch)
{
	return tolower(ch);
}

unsigned char glk_char_to_upper(unsigned char ch)
{
	return toupper(ch);
}

winid_t glk_window_get_root(void)
{
	return rootwin;
}


// Recursive resize
static void window_resize(winid_t root)
{
	winid_t keywin = root->keywin;
	int method = root->method;
	int splitm = method & winmethod_DirMask;
	int size;

	winid_t win0 = root->kids[0];
	winid_t win1 = root->kids[1];

	if(root->console_win >= 0)
	{
		setwin(root);
		console_ioctl(1, CC_MOVEWIN, root->x, root->y, root->width, root->height);
	}

	if(!(win0 && win1))
		return;

	if(keywin->wintype == wintype_TextBuffer)
	{
		font_width = type_fonts[1]->charwidth + type_fonts[1]->spacing;
		font_height = type_fonts[1]->height;
	}
	else
	if(keywin->wintype == wintype_Graphics)
	{
		font_width = font_height = 1;
	}
	else
	//if(win1->wintype == wintype_TextGrid)
	{
		font_width = type_fonts[0]->charwidth + type_fonts[1]->spacing;
		font_height = type_fonts[0]->height;
	}

	if(method & winmethod_Proportional)
	{
		if(splitm == winmethod_Left || splitm == winmethod_Right)
			size = root->width * root->size / 100;
		else
			size = root->height * root->size / 100;
	}
	else
	{
		if(splitm == winmethod_Left || splitm == winmethod_Right)
			size = font_width * root->size;
		else
			size = font_height * root->size;
	}

	win0->x = win1->x = root->x;
	win0->y = win1->y = root->y;
	win0->width = win1->width = root->width;
	win0->height = win1->height = root->height;

	if(splitm == winmethod_Left || splitm == winmethod_Right) // Vertical split
	{
		win1->width = size;
		win0->width = root->width - size;
		if(splitm == winmethod_Left)
			win0->x += size;
		else
			win1->x += size;
	}
	else // Horizontal split
	{
		win1->height = size;
		win0->height = root->height - size;
		if(splitm == winmethod_Above)
			win0->y += size;
		else
			win1->y += size;
	}

	// Resize/move old window
	if(win0->console_win >= 0)
	{
		setwin(win0);
		console_ioctl(1, CC_MOVEWIN, win0->x, win0->y, win0->width, win0->height);
	}
	if(win1->console_win >= 0)
	{
		setwin(win1);
		console_ioctl(1, CC_MOVEWIN, win1->x, win1->y, win1->width, win1->height);
	}

	window_resize(win0);
	window_resize(win1);
}

// Splits window, returns new parent with first kid as old parent
static winid_t window_split(winid_t split, int method, int size)
{
	winid_t paired = (winid_t)malloc(sizeof(struct glk_window_struct));
	winid_t win = (winid_t)malloc(sizeof(struct glk_window_struct));

	// Change tree so parent->split becomes parent --> paired -->split
	//                                                        '->win
	paired->parent = split->parent;
	paired->kids[0] = split;
	paired->kids[1] = win;

	if(split->parent)
	{
		if(split->parent->kids[0] == split)
			split->parent->kids[0] = paired;
		else
			split->parent->kids[1] = paired;
	}

	split->parent = paired;
	win->parent = paired;

	win->kids[0] = win->kids[1] = NULL;

	if(split == rootwin)
		rootwin = paired;
	// Tree should be correct at this point



	paired->console_win = -1;

	
	paired->wintype = wintype_Pair;
	paired->size = size;
	paired->method = method;
	paired->x = split->x;
	paired->y = split->y;
	paired->width = split->width;
	paired->height = split->height;
	paired->console_win = -1;
	paired->stream = paired->echo = NULL;

	paired->keywin = win;

    if (register_dispatch)
	{
        paired->disprock = (*register_dispatch)(paired, gidisp_Class_Window);
        win->disprock = (*register_dispatch)(win, gidisp_Class_Window);
	}
	else
	{
		paired->disprock.ptr = NULL;
		win->disprock.ptr = NULL;
	}
		
	return paired;	
}

#define WIN_GRAPHICS 3

int fg_colors[2] = { CCOL_WHITE, CCOL_BLACK };
int bg_colors[2] = { CCOL_BLUE, CCOL_WHITE };

char *SPLITDIR[] = { "winmethod_Left", "winmethod_Right", "winmethod_Above", "winmethod_Below" };
char *SPLITDIV[] = { "winmethod_Fixed", "winmethod_Proportional" };
char *WINTYPE[] = { "wintype_AllTypes", "wintype_Pair", "wintype_Blank", "wintype_TextBuffer", "wintype_TextGrid", "wintype_Graphics" };

//#include "windows.h"

int WinExists(winid_t win)
{
	winid_t w;
	if(win == NULL)
		return 1;

	w = glk_window_iterate(NULL, NULL);
	while(w)
	{
		if(w == win)
			return 1;
		w = glk_window_iterate(w, NULL);
	}

		return 0;

}

winid_t glk_window_open(winid_t split, glui32 method, glui32 size, glui32 wintype, glui32 rock)
{
	int i;
	int type = WIN_GRAPHICS;
	Font *fonts[4];// = {0, 0, 0, 0};
	//char tmp[128];
	//sprintf(tmp, "GLKCALL: glk_window_open(%p, %s %s, %d, %s, %d)\n", split, SPLITDIR[method & 0xf], SPLITDIV[(method>>4)-1],
	//        size, WINTYPE[wintype], rock);
	//OutputDebugStringA(tmp);

	assert(WinExists(split));

	for(i=0; i<4; i++)
		fonts[i] = 0;//type_fonts[i];

	if(wintype == wintype_TextBuffer)
	{
		for(i=0; i<4; i++)
			fonts[i] = type_fonts[i];
		type = CONWIN_STREAM;
	}
	else
	if(wintype == wintype_TextGrid)
	{
		fonts[0] = type_fonts[0];
		type = CONWIN_FIXED;
	}

	if(!split)
	{
		// Main window
		rootwin = (winid_t)malloc(sizeof(struct glk_window_struct));
		rootwin->parent = rootwin->kids[0] = rootwin->kids[1] = NULL;

	    if (register_dispatch)
	        rootwin->disprock = (*register_dispatch)(rootwin, gidisp_Class_Window);
		else
			rootwin->disprock.ptr = NULL;


		rootwin->console_win = -1;
		rootwin->stream = NULL;

		if(wintype != wintype_Graphics)
		{
			//fonts[0] = type_fonts[type];
			rootwin->console_win = console_ioctl(1, CC_OPENWIN, 0, 0, screen_width, screen_height, fonts, type);
			console_ioctl(1, CC_SETFONT, type);
			console_ioctl(1, CC_SETRGB, winstyles[wintype].fg[0], winstyles[wintype].bg[0]);
			console_ioctl(1, CC_CLEAR);
			rootwin->stream = new_stream(0);
			rootwin->stream->window = rootwin;


		}

		rootwin->wintype = wintype;
		rootwin->rock = rock;
		rootwin->x = rootwin->y = 0;
		rootwin->width = screen_width;
		rootwin->height = screen_height;
		
		rootwin->method = method;
		rootwin->size = size;

		rootwin->keywin = NULL;

		return rootwin;
	}
	else
	{
		// First we create a new paired window and a new empty window
		winid_t parent = window_split(split, method, size);

		// parent should now have 'split' as its first child and the new window as its second

		//winid_t oldwin = parent->kids[0];
		winid_t newwin = parent->kids[1];

		// Fill in as much as we can about the new window
		newwin->wintype = wintype;
		newwin->rock = rock;
		newwin->stream = NULL;

		if(wintype != wintype_Graphics)
		{
			newwin->stream = new_stream(0);
			newwin->stream->window = newwin;
		}

		newwin->console_win = -1; // Will prevent physical console window resize

		// Will resize the old and new window 
		window_resize(parent);

		// Open new window
		if(wintype != wintype_Graphics)
		{
			//fonts[0] = type_fonts[type];
			//fonts[1] = type_fonts[1];
			newwin->console_win = console_ioctl(1, CC_OPENWIN, newwin->x, newwin->y, newwin->width, newwin->height, fonts, type);
			console_ioctl(1, CC_SETFONT, type);
			//console_ioctl(1, CC_SETCOLOR, fg_colors[type], bg_colors[type]);
			console_ioctl(1, CC_SETRGB, winstyles[wintype].fg[0], winstyles[wintype].bg[0]);
			console_ioctl(1, CC_CLEAR);
		}

		//glk_set_window(newwin);

		return newwin;
	}
}

static void wclose(winid_t win)
{

    if (unregister_dispatch)
	{
        (*unregister_dispatch)(win, gidisp_Class_Window, win->disprock);
        win->disprock.ptr = NULL;
    }

	if(win->kids[0])
		wclose(win->kids[0]);
	if(win->kids[1])
		wclose(win->kids[1]);

	if(win->console_win >= 0)
		console_ioctl(1, CC_CLOSEWIN, win->console_win);
	if(win->stream)
		free_stream(win->stream);

//	free(win);
}

void glk_window_close(winid_t win, stream_result_t *result)
{
	winid_t pp, sibling;
	winid_t parent = win->parent;
	//char tmp[128];

	assert(WinExists(win));

	//sprintf(tmp, "GLKCALL: glk_window_close(%p)\n", win);
	//OutputDebugStringA(tmp);


	if(win == linewin)
		linewin = NULL;

	if(!parent)
	{
		wclose(win);
		rootwin = NULL;
		return;
	}

	if(parent->kids[0] == win)
		sibling = parent->kids[1];
	else
		sibling = parent->kids[0];

	sibling->x = parent->x;
	sibling->y = parent->y;
	sibling->width = parent->width;
	sibling->height = parent->height;

	if((pp = parent->parent))
	{
		if(pp->kids[0] == parent)
			pp->kids[0] = sibling;
		else
			pp->kids[1] = sibling;
	}
	else
	{
		rootwin = sibling;
	}

	sibling->parent = pp;


	wclose(win);
	window_resize(sibling);

	parent->kids[0] = parent->kids[1] = NULL;
	wclose(parent);
}

void glk_window_get_size(winid_t win, glui32 *widthptr, glui32 *heightptr)
{
	int w = 80;
	int h = 24;
	
	/*if(win->wintype == wintype_TextBuffer)
	{

		font_width = type_fonts[1]->charwidth + type_fonts[1]->spacing;
		font_height = type_fonts[1]->height;
	}
	else
	if(win->wintype == wintype_TextGrid)
	{
		font_width = type_fonts[0]->charwidth + type_fonts[0]->spacing;
		font_height = type_fonts[0]->height;
	}
	else
	if(win->wintype == wintype_Graphics)
	{
		font_width = font_height = 1;
	}
	else
		return;
*/

	if(win->wintype == wintype_Graphics)
	{
		w = win->width;
		h = win->height;
	}
	else
	{
		setwin(win);
		w = console_ioctl(1, CC_GETWIDTH);
		h = console_ioctl(1, CC_GETHEIGHT);
	}

	// Needed to start trinity, but will mess upp centering
	//if(*widthptr < 62)
	//	*widthptr = 62;

	if(widthptr)
		*widthptr = w;
	if(heightptr)
		*heightptr = h;
}

void glk_window_set_arrangement(winid_t win, glui32 method, glui32 size, winid_t keywin)
{
	//char tmp[128];
	assert(win->wintype == wintype_Pair);
	assert(WinExists(win));
	assert(win->kids[0] && win->kids[1]);

	//assert(win->keywin == keywin);
	if(keywin)
	{
		if(win->kids[0] == keywin)
		{
			win->kids[0] = win->kids[1];
			win->kids[1] = keywin;
		}
		win->keywin = keywin;
	}

	//sprintf(tmp, "GLKCALL: glk_window_set_arrangement(%p, %s %s, %d, %p)\n", win, SPLITDIR[method & 0xf], SPLITDIV[(method>>4)-1],
	//        size, keywin);
	//OutputDebugStringA(tmp);


	win->method = method;
	win->size = size;
	window_resize(win);
}

void glk_window_get_arrangement(winid_t win, glui32 *methodptr, glui32 *sizeptr, winid_t *keywinptr)
{
	if(methodptr)
		*methodptr = win->method;
	if(sizeptr)
		*sizeptr = win->size;
	if(keywinptr)
		*keywinptr = win->keywin;
}

winid_t glk_window_iterate(winid_t win, glui32 *rockptr)
{
	if(win)
	{
		if(win->kids[0])
			return win->kids[0];
		else
		{
			while(win->parent && win->parent->kids[1] == win)
				win = win->parent;

			if(!win->parent)
				return NULL;

			return win->parent->kids[1];
		}
	}
	return rootwin;
}

glui32 glk_window_get_rock(winid_t win)
{
	return win->rock;
}

glui32 glk_window_get_type(winid_t win)
{
	return win->wintype;
}

winid_t glk_window_get_parent(winid_t win)
{
	return win->parent;
}

winid_t glk_window_get_sibling(winid_t win)
{
	if(win == win->parent->kids[0])
		return win->parent->kids[1];
	else
		return win->parent->kids[0];
}

void glk_window_clear(winid_t win)
{
	if(win->wintype == wintype_Graphics)
	{
		uint16 bg = 0xFFFF;
		int w;
		int h = win->height;
		uint16 *ptr = &((uint16 *)screen_pixels)[win->x + win->y * screen_width];
		
		while(h--)
		{
			w = win->width;
			while(w--)
				*ptr++ = bg;

			ptr += (screen_width - win->width);
		}

	}
	else
	{
		setwin(win);
		console_ioctl(1, CC_CLEAR);
	}
}

void glk_window_move_cursor(winid_t win, glui32 xpos, glui32 ypos)
{
	setwin(win);
	console_ioctl(1, CC_GOTOXY, xpos, ypos);
}


strid_t glk_window_get_stream(winid_t win)
{
	return win->stream;
}

void glk_window_set_echo_stream(winid_t win, strid_t str)
{
	win->echo = str;
}

strid_t glk_window_get_echo_stream(winid_t win)
{
	return win->echo;
}

void glk_set_window(winid_t win)
{
	active_stream = win->stream;
}

strid_t glk_stream_open_file(frefid_t fileref, glui32 fmode, glui32 rock)
{
	strid_t str = new_stream(rock);

//	dprintf("glk_stream_open_file(%s)", fileref->name);

	if(fmode & filemode_Write)
		str->fp = pogo_fopen(fileref->name, "wb");
	else
		str->fp = pogo_fopen(fileref->name, "rb");
	
	return str;
}

strid_t glk_stream_open_memory(char *buf, glui32 buflen, glui32 fmode, glui32 rock)
{
	strid_t str = new_stream(rock);
	str->mem = buf;
	str->memlen = buflen;
	return str;
}

void glk_stream_close(strid_t str, stream_result_t *result)
{
	if(str->fp)
	{
		if(result)
			result->writecount = result->readcount = pogo_ftell(str->fp);
		pogo_fclose(str->fp);
	}
	else if(str->mem)
	{
		if(result)
			result->writecount = result->readcount = str->mempos;
	}

}

strid_t glk_stream_iterate(strid_t str, glui32 *rockptr)
{
	return NULL;
}

glui32 glk_stream_get_rock(strid_t str)
{
	return str->rock;
}

void glk_stream_set_position(strid_t str, glsi32 pos, glui32 seekmode)
{
	int mode = 0;
	switch(seekmode)
	{
	case seekmode_Start : mode = SEEK_SET; break;
	case seekmode_Current : mode = SEEK_CUR; break;
	case seekmode_End : mode = SEEK_END; break;
	}
	if(str->fp)
	{
		pogo_fseek(str->fp, pos, mode);
	}
	else if(str->mem)
	{
		switch(mode)
		{
		case SEEK_SET:
			str->mempos = pos;
			break;
		case SEEK_CUR:
			str->mempos += pos;
			break;
		case SEEK_END:
			str->mempos = str->memlen - pos;
			break;
		}
		
		if(str->mempos < 0)
			str->mempos = 0;
		if(str->mempos > str->memlen)
			str->mempos = str->memlen;
	}
}

glui32 glk_stream_get_position(strid_t str)
{
	if(str->fp)
	{
		return pogo_ftell(str->fp);
		//return fseek(str->fp, 0, SEEK_CUR);
	}
	else if(str->mem)
	{
		return str->mempos;
	}

	return -1;
}

void glk_stream_set_current(strid_t str)
{
	active_stream = str;
}

strid_t glk_stream_get_current(void)
{
	return active_stream;
}


void glk_put_char(unsigned char ch)
{
	glk_put_char_stream(active_stream, ch);
}



void glk_put_char_stream(strid_t str, unsigned char ch)
{
	if(str->window)
	{
		int x, y;
		console_ioctl(1, CC_SETWIN, str->window->console_win);
		x = console_ioctl(1, CC_GETX);
		y = console_ioctl(1, CC_GETY);

		if(str->window->wintype == wintype_TextGrid)
		{
			//char tmp[128];
			//if(ch != ' ')
			//{
			//	sprintf(tmp, "putchar (%02d,%02d) '%c' \n", x, y, ch);
			//	OutputDebugStringA(tmp);
			//}
		}

		console_write(1, &ch, 1);
	}
	else
	if(str->fp)
	{
		pogo_fputc(ch, str->fp);
	}
	else if(str->mem)
	{
		if(str->mempos < str->memlen)
			str->mem[str->mempos++] = ch;
	}

}


void glk_put_string(char *s)
{
	glk_put_buffer_stream(active_stream, s, strlen(s));
}

void glk_put_string_stream(strid_t str, char *s)
{
	glk_put_buffer_stream(str, s, strlen(s));
}

void glk_put_buffer(char *buf, glui32 len)
{
	glk_put_buffer_stream(active_stream, buf, len);
}

void glk_put_buffer_stream(strid_t str, char *buf, glui32 len)
{
	if(str->window)
	{
		console_ioctl(1, CC_SETWIN, str->window->console_win);
		console_write(1, buf, len);
	}
	if(str->fp)
	{
		pogo_fwrite(buf, 1, len, str->fp);
	}
	else if(str->mem)
	{
		glui32 left = str->memlen - str->mempos;
		if(left < len)
			len = left;

		if(len)
		{
			memcpy(&str->mem[str->mempos], buf, len);
			str->mempos += len;
		}
	}
}

int lastf = CCOL_BLACK;
int lastb = CCOL_WHITE;

void glk_set_colors(strid_t str, glui32 fg, glui32 bg)
{
	if(fg == 1)
		fg = CCOL_WHITE;
	else
	if(fg == 0)
		fg = lastf;
	else
	if(fg >= 10)
		fg = CCOL_GRAY;
	else
		fg += 6;

	if(bg == 1)
		bg = CCOL_BLUE;
	else
	if(bg == 0)
		bg = lastb;
	else
	if(bg >= 10)
		bg = CCOL_GRAY;
	else
		bg -= 2;

	lastf = fg;
	lastb = bg;

	if(str->window)
	{
		setwin(str->window);
		console_ioctl(1, CC_SETCOLOR, fg, bg);
	}
}

void glk_set_style(glui32 styl)
{
	glk_set_style_stream(active_stream, styl);
}

void glk_set_style_stream(strid_t str, glui32 styl)
{
	winid_t win;
	if(!str)
		win = rootwin;
	else
		win = str->window;

	if(!win || win->wintype != wintype_TextBuffer)
		return;

	setwin(win);

	switch(styl)
	{
	case style_User2:
	case style_Emphasized:
		console_ioctl(1, CC_SETFONT, FONT_ITALIC);
		break;
	case style_User1:
	case style_Subheader:
		console_ioctl(1, CC_SETFONT, FONT_BOLD);
		break;
	case style_Preformatted:
		console_ioctl(1, CC_SETFONT, FONT_FIXED);
		break;
	default:
		console_ioctl(1, CC_SETFONT, FONT_NORMAL);
		break;
	}

}

glsi32 glk_get_char_stream(strid_t str)
{
	if(str->fp)
	{
		int c = pogo_fgetc(str->fp);
		return c;
	}
	else if(str->mem)
	{
		if(str->mempos < str->memlen)
			return (glsi32)str->mem[str->mempos++];
	}
	return -1;
}

glui32 glk_get_line_stream(strid_t str, char *buf, glui32 len)
{
	if(str->fp)
	{
		int l;
		if(fgets(buf, len, str->fp))
		{
			l = strlen(buf);
			buf[l-1] = 0;
			return l-1;
		}
	}
	return 0;
}

glui32 glk_get_buffer_stream(strid_t str, char *buf, glui32 len)
{
	if(str->fp)
	{
		return pogo_fread(buf, 1, len, str->fp);
	}
	else if(str->mem)
	{
		glui32 left = str->memlen - str->mempos;
		if(left < len)
			len = left;
		if(len)
		{
			memcpy(buf, &str->mem[str->mempos], len);
			str->mempos += len;
		}

		return len;
	}

	return 0;
}



void glk_stylehint_set(glui32 wintype, glui32 styl, glui32 hint, glsi32 val)
{
	int i;
	switch(hint)
	{
		case stylehint_TextColor:
			if(wintype)
				winstyles[wintype].fg[styl] = val;
			else
				for(i=0; i<6; i++)
					winstyles[i].fg[styl] = val;
			break;
		case stylehint_BackColor:
			if(wintype)
				winstyles[wintype].bg[styl] = val;
			else
				for(i=0; i<6; i++)
					winstyles[i].bg[styl] = val;
			break;
	}	
}

void glk_stylehint_clear(glui32 wintype, glui32 styl, glui32 hint)
{
}

glui32 glk_style_distinguish(winid_t win, glui32 styl1, glui32 styl2)
{
	return 1;
}

glui32 glk_style_measure(winid_t win, glui32 styl, glui32 hint, glui32 *result)
{
	return 0;
}



frefid_t glk_fileref_create_temp(glui32 usage, glui32 rock)
{

	return 0;
}

frefid_t glk_fileref_create_by_name(glui32 usage, char *name, glui32 rock)
{
//#ifdef POGO
//	const char prefix[] = "/rom/";
//#else
//	const char prefix[] = "data/";
//#endif
	const char prefix[] = "";

	frefid_t f = new_fileref(rock);
	f->name = malloc(strlen(name) + sizeof(prefix) + 1);
	//f->name = name;
	sprintf(f->name, "%s%s", prefix, name);
	//f->fp = fopen(name, "rb");
	//f->stream = new_stream(rock);
	return f;
}

char save_filename[64] = "game.sav";
char transcript_filename[64] = "game.trans";
char data_filename[64] = "game.dat";

//#ifdef _WIN32
//char dirname[128] = "c:\\DATA\\IF\\Games\\";
//#else
char dirname[128] = "/";// /disk/";
//#endif

frefid_t glk_fileref_create_by_prompt(glui32 usage, glui32 fmode, glui32 rock)
{
	frefid_t f;
	//char fname[128];
	char *filename, *txt;
	/*
	if(!strlen(dirname))
	{
		struct pogo_stat st;
		strcpy(dirname, "/disk");
		getcwd(&dirname[5], sizeof(dirname));
		if(pogo_stat(dirname, &st) != 0)
			dirname[5] = 0;
		else
			strcat(dirname, "/");
	}
	*/
	f = new_fileref(rock);
	f->mode = fmode;

	switch(usage)
	{
	case fileusage_SavedGame:
		txt = (fmode & filemode_Read) ? "Restore Progress" : "Save Progress";
		filename = save_filename;
		break;
	case fileusage_Transcript:
		txt = (fmode & filemode_Read) ? "Load Transcript" : "Save Transcript";
		filename = transcript_filename;
		break;
	default:
		txt = (fmode & filemode_Read) ? "Load file" : "Save file";
		filename = data_filename;
		break;
	}

	if(file_request(txt, dirname, 128, filename, 128))
	{
		f->name = malloc(strlen(dirname)+strlen(filename)+1);
		strcpy(f->name, dirname);
		strcat(f->name, filename);
		return f;
	}

	free(f);
	return NULL;

	//return glk_fileref_create_by_name(usage, name, rock);
}

frefid_t glk_fileref_create_from_fileref(glui32 usage, frefid_t fref, glui32 rock)
{
	UNIMP(0);
}

void glk_fileref_destroy(frefid_t fref)
{
	if(fref->name)
		free(fref->name);
	
	free(fref);

	//UNIMP();
}

frefid_t glk_fileref_iterate(frefid_t fref, glui32 *rockptr)
{
	UNIMP(0);
}

glui32 glk_fileref_get_rock(frefid_t fref)
{
	UNIMP(0);
}

void glk_fileref_delete_file(frefid_t fref)
{
	//UNIMP(0);
}

glui32 glk_fileref_does_file_exist(frefid_t fref)
{
	UNIMP(1);
}

unsigned int timer_interval = 0;
unsigned int last_time = 0;

void glk_select(event_t *event)
{
	unsigned int ticks;
	event_t tmp;

	//glk_put_char_stream(bufstr, 0);
	console_ioctl(1, CC_FLUSH);
	graphics_flush();

	event->val1 = 0;
	event->val1 = 0;

	if(do_redraw)
	{
		event->type = evtype_Redraw;
		event->win = rootwin;
		do_redraw = 0;
		return;
	}


	while(1)
	{
		int rc;
		ticks = events_getmsec();
		if(timer_interval && ((ticks - last_time ) >= timer_interval))
		{
			event->type = evtype_Timer;
			last_time = ticks;
			return;
		}

		rc = events_get(&tmp);
		if(rc == -1)
			glk_exit();
		if(rc)
		{
			if(tmp.type == evtype_Arrange)
			{
				if(linewin && (listen_flags & LINE_EVENT))
				{
					setwin(linewin);
					readline_end(1);
					rearrange();
					setwin(linewin);
					readline_start(200);
					graphics_flush();
				}
				else
					rearrange();

			}
			else
			if(tmp.type == evtype_CharInput)
			{
				console_ioctl(1, CC_RESETMORE);
				
			
				if(linewin && (listen_flags & CHAR_EVENT))
				{
					event->type = evtype_CharInput;
					event->val1 = tmp.val1;
					event->win = linewin;
					listen_flags &= ~CHAR_EVENT;
					return;
				}
				else
				if(linewin && (listen_flags & LINE_EVENT))
				{
					if(tmp.val1 == keycode_Return)
					{
						const char cr = '\n';
						strcpy(linebuf, readline_end(0));
						event->type = evtype_LineInput;
						event->val1 = strlen(linebuf);
						event->win = linewin;
						listen_flags &= ~LINE_EVENT;
						setwin(linewin);
						console_write(1, &cr, 1);
						/*{
							char tmp[64];
							sprintf(tmp, "'%s'\n", linebuf);
							console_write(1, tmp, strlen(tmp));
						}*/
						//console_write(1, &cr, 1);
						graphics_flush();
						return;
					}
					else
					{
						int v = tmp.val1;
						setwin(linewin);

						switch(v)
						{
						case keycode_Left: v = RL_KEY_LEFT; break;
						case keycode_Right: v = RL_KEY_RIGHT; break;
						case keycode_Up: v = RL_KEY_UP; break;
						case keycode_Down: v = RL_KEY_DOWN; break;
						case keycode_Home: v = RL_KEY_HOME; break;
						case keycode_End: v = RL_KEY_END; break;
						case keycode_Delete: v = RL_KEY_DEL; break;
						case keycode_Tab: v = RL_KEY_COMPLETE; break;
						case keycode_Func12: v = RL_KEY_KILL; break;
						}

						if(linewin->wintype == wintype_TextBuffer)
							console_ioctl(1, CC_SETFONT, 3);

						if(readline_putchar(v))
							graphics_flush();

						if(linewin->wintype == wintype_TextBuffer)
							console_ioctl(1, CC_SETFONT, 1);
					}
				}
			}
		}
	}
}

void glk_select_poll(event_t *event)
{
}


void glk_request_timer_events(glui32 millisecs)
{
	timer_interval = millisecs;
	last_time = events_getmsec();
}

void glk_request_line_event(winid_t win, char *buf, glui32 maxlen, glui32 initlen)
{
	if(linewin)
	{
		setwin(linewin);
		console_ioctl(1, CC_MARK, -1, -1);
	}

	linebuf = buf;
	linewin = win;
	listen_flags = LINE_EVENT;
	setwin(win);
	readline_start(200);
	graphics_flush();
}

void glk_request_char_event(winid_t win)
{
	if(linewin)
	{
		setwin(linewin);
		console_ioctl(1, CC_MARK, -1, -1);
	}
	listen_flags = CHAR_EVENT;
	linewin = win;
}

void glk_request_mouse_event(winid_t win)
{
}


void glk_cancel_line_event(winid_t win, event_t *event)
{
	if(linebuf)
		strcpy(linebuf, readline_end(0));
	if(event)
	{
		event->type = evtype_LineInput;
		event->val1 = strlen(linebuf);
		event->val2 = 0;
	}
	listen_flags &= ~LINE_EVENT;
	linewin = NULL;
}

void glk_cancel_char_event(winid_t win)
{
	listen_flags &= ~CHAR_EVENT;
}

void glk_cancel_mouse_event(winid_t win)
{
}

static giblorb_map_t *blorbmap = 0;

giblorb_err_t giblorb_set_resource_map(strid_t file)
{
	giblorb_err_t err;

	err = giblorb_create_map(file, &blorbmap);
	if (err)
	{
		blorbmap = 0;
		return err;
	}

	return giblorb_err_None;
}

giblorb_map_t *giblorb_get_resource_map()
{
	return blorbmap;
}

#ifdef GLK_MODULE_IMAGE

#define imagealign_InlineUp (0x01)
#define imagealign_InlineDown (0x02)
#define imagealign_InlineCenter (0x03)
#define imagealign_MarginLeft (0x04)
#define imagealign_MarginRight (0x05)

void *img_data;
glui32 img_len;
int img_id;
int img_no = -1;

void load_image(int n)
{
	giblorb_result_t Result;
	if (giblorb_load_resource(blorbmap, giblorb_method_Memory, &Result, giblorb_ID_Pict, n) == giblorb_err_None)
	{
		img_data = Result.data.ptr;
		img_len = Result.length;
		img_id = Result.chunktype;
		img_no = n;
	}
	
}

glui32 glk_image_draw(winid_t win, glui32 image, glsi32 val1, glsi32 val2)
{
	return 0;
}

glui32 glk_image_draw_scaled(winid_t win, glui32 image, glsi32 val1, glsi32 val2, glui32 width, glui32 height)
{
	return 0;
}

glui32 glk_image_get_info(glui32 image, glui32 *width, glui32 *height)
{

	return 0;
}

void glk_window_flow_break(winid_t win)
{
}


void glk_window_erase_rect(winid_t win, glsi32 left, glsi32 top, glui32 width, glui32 height)
{
}

void glk_window_set_background_color(winid_t win, glui32 color)
{
}


//#define UINT2PIXEL(x) ( (uint16)( ((x & 0xF80000)>>8) | ((x & 0xFC00)>>5)  | ((x & 0xF8)>>3) ) )

void glk_window_fill_rect(winid_t win, glui32 color, glsi32 left, glsi32 top, glui32 width, glui32 height)
{
		//uint16 bg = 0;
		int w,h;
		uint16 col16 = UINT2PIXEL(color);
		uint16 *ptr;

		if(top >= win->height || left >= win->width)
			return;
		
		if(top<0)
			top = 0;

		if(left<0)
			left = 0;

		if((left+width) > win->width)
			width = (win->width - left);
		
		if((top+height) > win->height)
			height = (win->height - top);
	
		h = height;

		ptr = &((uint16 *)screen_pixels)[win->x + left + (win->y + top) * screen_width];

		while(h--)
		{
			w = width;
			while(w--)
				*ptr++ = col16;
			ptr += (screen_width - width);
		}
}


#endif /* GLK_MODULE_IMAGE */

#ifdef GLK_MODULE_SOUND

schanid_t glk_schannel_create(glui32 rock);
void glk_schannel_destroy(schanid_t chan); 
schanid_t glk_schannel_iterate(schanid_t chan, glui32 *rockptr);
glui32 glk_schannel_get_rock(schanid_t chan);

glui32 glk_schannel_play(schanid_t chan, glui32 snd);
glui32 glk_schannel_play_ext(schanid_t chan, glui32 snd, glui32 repeats,
    glui32 notify);
void glk_schannel_stop(schanid_t chan);
void glk_schannel_set_volume(schanid_t chan, glui32 vol);

void glk_sound_load_hint(glui32 snd, glui32 flag);

#endif /* GLK_MODULE_SOUND */

#ifdef GLK_MODULE_HYPERLINKS

void glk_set_hyperlink(glui32 linkval);
void glk_set_hyperlink_stream(strid_t str, glui32 linkval);
void glk_request_hyperlink_event(winid_t win);
void glk_cancel_hyperlink_event(winid_t win);

#endif

void gidispatch_set_object_registry(gidispatch_rock_t (*regi)(void *, glui32), void (*unregi)(void *, glui32, gidispatch_rock_t))
{ 
    winid_t win;
    strid_t str;
    frefid_t fref;

	register_dispatch = regi;
	unregister_dispatch = unregi;

    if (register_dispatch) {
        /* It's now necessary to go through all existing objects, and register
            them. */
        for (win = glk_window_iterate(NULL, NULL); 
            win;
            win = glk_window_iterate(win, NULL)) {
            win->disprock = (*register_dispatch)(win, gidisp_Class_Window);
        }
        for (str = glk_stream_iterate(NULL, NULL); 
            str;
            str = glk_stream_iterate(str, NULL)) {
            str->disprock = (*register_dispatch)(str, gidisp_Class_Stream);
        }
        for (fref = glk_fileref_iterate(NULL, NULL); 
            fref;
            fref = glk_fileref_iterate(fref, NULL)) {
            fref->disprock = (*register_dispatch)(fref, gidisp_Class_Fileref);
        }
    }



}

gidispatch_rock_t gidispatch_get_objrock(void *obj, glui32 objclass)
{
    switch (objclass)
	{
        case gidisp_Class_Window:
            return ((winid_t)obj)->disprock;
        case gidisp_Class_Stream:
            return ((strid_t)obj)->disprock;
        case gidisp_Class_Fileref:
            return ((frefid_t)obj)->disprock;
        default:
		{
            gidispatch_rock_t dummy;
            dummy.num = 0;
            return dummy;
        }
    }
}
