
#include <pogo/pogo.h>
#include <pogo/ds_defs.h>
#include "glk.h"

void debug_print(const char *sz)
{
/*	asm volatile(
	"mov r0, %0\n"
	"swi 0xff0000\n"
  :
  :
  "r" (sz) :
  "r0", "r1", "r2");*/
}


int dprintf(char *fmt, ...)
{
	static char tmp[256];
	int rc;
	va_list vl;
	va_start(vl, fmt);

	rc = vsprintf(tmp, fmt, vl);
	debug_print((const char *)tmp);
	va_end(vl);
	return rc;
}


void screen_init();

void clear_screen(uint16 col)
{
	int l = 256*192;
	uint16 *ptr = (uint16*)0x06800000;
	while(l--)
		*ptr++ = col;
}


int graphics_init(void **pixels, int *width, int *height, int *stride)
{
	*pixels = (void *)0x06800000;
	*width = *stride = 256;
	*height = 192;


	clear_screen(0x03C0);

	screen_init();
		
	clear_screen(0x3C00);

	return 1;

}

void graphics_flush()
{
}

void graphics_exit()
{
}

extern int frame_counter;

int events_getmsec()
{
	return frame_counter * 20;
}

static uint16 oldkeys;

int input_mode = 0;

int events_get(event_t *glkevt)
{
	uint16 keys;
	int c;
	
	keys = ~(*KEYS);
		
	if(keys != oldkeys)
	{
		oldkeys = keys;

		glkevt->type = evtype_CharInput;
		glkevt->val1 = keycode_Unknown;

		if(keys & 0x10)
			glkevt->val1 = keycode_Right;
		else
		if(keys & 0x20)
			glkevt->val1 = keycode_Left;
		else
		if(keys & 0x80)
			glkevt->val1 = keycode_Down;
		else
		if(keys & 0x40)
			glkevt->val1 = keycode_Up;
		else
		if(keys & 0x1)
			glkevt->val1 = ' ';
		else
		if(keys & 0x2)
			glkevt->val1 = keycode_Delete;
		else
		if(keys & 0x8)
			glkevt->val1 = keycode_Return;
		else
		if(keys & 0x4)
		{
			//glkevt->val1 = keycode_Escape;
			input_mode ^= 1;
			vkey_mode(input_mode);
			if(input_mode)
				*POWER_CR = POWER_ALL_2D;// | 0x8000;
			else
				*POWER_CR = POWER_ALL_2D | 0x8000;
			return 0;
		}
		else
		/*if(keys & 0x100)
			glkevt->val1 = keycode_End;
		else
		if(keys & 0x200)
			glkevt->val1 = keycode_Home;
		else*/
			return 0;

		dprintf("%d", glkevt->val1);

		return 1;
	}

	if((c = pogo_getchar()) != EOF)
	{
		glkevt->type = evtype_CharInput;
		glkevt->val1 = keycode_Unknown;


		switch(c)
		{
		/*case SDLK_RETURN: glkevt->val1 = keycode_Return;
			break;
		case SDLK_RIGHT: glkevt->val1 = keycode_Right;
			break;
		case SDLK_LEFT: glkevt->val1 = keycode_Left;
			break;
		case SDLK_UP: glkevt->val1 = keycode_Up;
			break;
		case SDLK_DOWN: glkevt->val1 = keycode_Down;
			break;
		case SDLK_HOME: glkevt->val1 = keycode_Home;
			break;
		case SDLK_END: glkevt->val1 = keycode_End;
			break;
		case SDLK_BACKSPACE: glkevt->val1 = keycode_Delete;
			break;
		case SDLK_ESCAPE: glkevt->val1 = keycode_Escape;
			break;
		case SDLK_PAGEUP: glkevt->val1 = keycode_PageUp;
			break;
		case SDLK_PAGEDOWN: glkevt->val1 = keycode_PageDown;
			break;*/
		case 2:
			input_mode ^= 1;
			vkey_mode(input_mode);
			if(input_mode)
				*POWER_CR = POWER_ALL_2D;// | 0x8000;
			else
				*POWER_CR = POWER_ALL_2D | 0x8000;
			return 0;
			break;
		case 8:
			glkevt->val1 = keycode_Delete;
			break;
		case 13:
		case 10:
			glkevt->val1 = keycode_Return;
			break;
		default:
			//if(event.key.keysym.unicode >= 0x20 && event.key.keysym.unicode <= 0xff)
			//	glkevt->val1 = event.key.keysym.unicode;
			//else
			//	return 0;
			if(c < 0x20 || c > 0x7e)
				c = 'X';
			glkevt->val1 = c;
			break;
		}



		dprintf("KEY %d", c);
		return 1;
	}

	return 0;
}

void vkey_init();
void vkey_generate();
void gbfsdev_init();
void fatdev_init();

void io_init()
{
	int i;
	char *layout2 = "1234567890qwertyuiopasdfghjkl;zxcvbnm,.EH#\"SSSSBBE";

	//FAT_InitFiles();

	debug_print("io_init()");

	pogo_device_init();

	for(i=0; i<30; i++)
		clear_screen(0x007C);

	//gbfs_root = find_first_gbfs_file(gbfs_space);

	//while(1);

	//fatdev_init();
	//clear_screen(0x00F0);
	//for(i=0; i<30; i++)
	//	clear_screen(0x007C);

	gbfsdev_init();

	vkey_init();
	Font *f = font_load("arial.font");
	if(!f)
		f = font_load("/rom/arial.font");
	if(!f)
		f = font_getdefault();

	if(!f)
	{
		clear_screen(0xFF0F);
		while(1);
	}
	vkey_generate(f, layout2, 10, 5, 25);

	int fd = pogo_open("/dev/vkey", 0);

	dprintf("%d is vkey\n", fd);



}
