#ifndef POGODEVICE_H
#define POGODEVICE_H

#include "stdcore.h"

#ifdef __cplusplus
extern "C" {
#endif

//#define POGO_PACKED __attribute__ ((packed))
#define POGO_PACKED

typedef struct
{
	char name[32];
	int size;
	int start;

} POGO_PACKED Romfile;

struct pogo_stat
{
  uchar		pst_dev;
  uint16	pst_ino;
  uint16	pst_mode;
  uchar		pst_nlink;
  uchar		pst_uid;
  uchar		pst_gid;
  uchar		pst_rdev;
  uint32	pst_size;
  uint32	pst_atime;
  uint32	pst_mtime;
  uint32	pst_ctime;
};


typedef struct {
	int (*open)(const char *name, int flags);
	int (*read)(int fd, void *dest, int size);
	int (*write)(int fd, const void *dest, int size);
	void (*close)(int fd);
	int (*lseek)(int fd, int offset, int origin);
	int (*ioctl)(int fd, int request, va_list vlist);
	int (*stat)(const char *path, struct pogo_stat *buffer);
	int (*remove)(const char *name);
	int (*rename)(const char *oldname, const char *newname);
	int (*dopen)(const char *name);
	int (*dread)(int fd, void *dest, int size);
	void (*dclose)(int fd);
	int (*mkdir)(const char *name);

} PogoDevice;

int pogo_device_register(PogoDevice *dev, char *name, void (*irqfunc)(void), int fd);

void pogo_device_init();

int pogo_open(const char *name, int flags);
int pogo_read(int fd, void *dest, int size);
int pogo_write(int fd, const void *dest, int size);
int pogo_close(int fd);
int pogo_lseek(int fd, int offset, int origin);
int pogo_tell(int fd);
int pogo_ioctl(int fd, int request, ...);
int pogo_remove(const char *name);
int pogo_stat(const char *name, struct pogo_stat *buffer);

int pogo_dopen(const char *name);
int pogo_dclose(int fd);
int pogo_dread(int fd, void *dest, int size);

int pogo_mkdir(const char *name);

void pogo_execv(const char *cmdname, const char *const *argv);
void pogo_execv_jump(const char *cmdname, const char *const *argv, void *jump);

uint32 pogo_time(void *t);

#define POGO_CLOCKS_PER_SEC 60

uint32 pogo_clock(void);

#define POGO_STDIN 0
#define POGO_STDOUT 1
#define POGO_STDERR 2

#define	PG_SEEK_SET	0	/* set file offset to offset */
#define	PG_SEEK_CUR	1	/* set file offset to current plus offset */
#define	PG_SEEK_END	2	/* set file offset to EOF plus offset */
#define PG_SEEK_MEM 3   /* Return memory position if a romfile */


#define PGS_IFDIR 0x1000

#define PGO_RDONLY	0x0001
#define PGO_WRONLY	0x0002
#define PGO_RDWR	0x0003
#define PGO_NBLOCK	0x0010
#define PGO_APPEND	0x0100
#define PGO_CREAT	0x0200
#define PGO_TRUNC	0x0400
#define PGO_NOWAIT	0x8000


/* ioctl commands - TODO: Many obsolete definitions, possbile clib collisions */

/* Standard ioctl commands */
#define IO_GETMODE 1
#define IO_SETMODE 2
#define IO_GETPARAM 3
#define IO_SETPARAM 4

/* Keyboard ioctl commands */
#define KC_SETIMAGE 9
#define KC_GETSCREENOFFSET 10
#define KC_ADDHISTORY 5
#define KC_SETDICT 6
#define KC_BINDKEY 7
#define KC_SETCOMPLETION 8

/* SRAM ioctl commands */
#define SR_SETUSER 20
#define SR_GETUSER 21
#define SR_FREESPACE 22


#define KM_RAW 1
#define KM_SHOWKBD 2
#define KM_BLOCKED 4
#define KM_ECHO 8

#define RAWKEY_L 1
#define RAWKEY_R 2
#define RAWKEY_UP 3
#define RAWKEY_DOWN 4
#define RAWKEY_LEFT 5
#define RAWKEY_RIGHT 6
#define RAWKEY_START 7
#define RAWKEY_SELECT 14
#define RAWKEY_A 15
#define RAWKEY_B 16
#define RAWKEY_X 17
#define RAWKEY_Y 18

#define POGOKEY_UP 3
#define POGOKEY_DOWN 4
#define POGOKEY_LEFT 5
#define POGOKEY_RIGHT 6

#define POGOKEY_BACKSPACE 8
#define POGOKEY_TAB 9
#define POGOKEY_ENTER 10
#define POGOKEY_ESCAPE 27
#define POGOKEY_PAGEUP 19
#define POGOKEY_PAGEDOWN 20
#define POGOKEY_HOME 21
#define POGOKEY_END 22
#define POGOKEY_INSERT 23
#define POGOKEY_DEL 24

#define POGOKEY_KILL 25

#define RAWKEY_RELEASE 0x80

#define CHAR_LEFT 241
#define CHAR_RIGHT 242
#define CHAR_LINEKILL 250

#ifdef __cplusplus
}
#endif


#endif
