
//#include <nds.h>
#include <pogo/device.h>

#include <string.h>
#include <stdio.h>
#include <stdlib.h>

int32 first_file();
int32 next_file(int pos);
int32 get_fileinfo(int offs, int32 *length, char *filename);
int32 find_file(const char *fname);
void dump_files();
void write_file(int file_pos, int offs, const uchar *src, int size);
void read_file(int file_pos, int offs, uchar *dst, int size);
void remove_file(int offs);
int create_file(const char *fname);
int bump_file(int offs);
void rinit(void *memory, int size, int format);
void truncate_file(int file_pos, int size);

//int dprintf(char *fmt, ...);

static PogoDevice rdev;

typedef struct
{
	int mem_offs;
	int pos;
	int length;

} RFile;

static RFile *open_files[32];
static int init_of = 0;

static RFile *bumpfile;

int cprintf2(const char *format,...);


int rdev_open(const char *name, int flags)
{
	RFile *r;
	int i;
	int length;
	int offs = find_file(name);

	if(!init_of)
	{
		memset(open_files, 0, sizeof(RFile*) * 32);
		init_of = 1;
	}

	//cprintf2("Open %s in %smode = %d\n", name, (flags & PGO_WRONLY) ? "write" : "read", offs);

	if((flags & PGO_WRONLY))
	{
		if(offs)
		{
			truncate_file(offs, 0);
		}
		else
		{
			offs = create_file(name);
		}
	}

	if(!offs)
		return -1;


	get_fileinfo(offs, &length, NULL);

	r = malloc(sizeof(RFile));

	r->mem_offs = offs;
	r->pos = 0;
	r->length = length;

	i = 0;
	while(open_files[i]) i++;
	open_files[i] = r;
	return i;
}

int rdev_read(int fd, void *dest, int size)
{
	RFile *f = open_files[fd];
	int left = f->length - f->pos;

	//printf("Reading %d bytes with %d bytes left\n", size, left);


	if(size > left)
		size = left;

	read_file(f->mem_offs, f->pos, dest, size);
	f->pos += size;

	return size;
}

int rdev_write(int fd, const void *src, int size)
{
	int length;
	RFile *f = open_files[fd];

	//printf("Writing %d bytes\n", size);

	if(f->pos + size > f->length)// && (f != bumpfile))
	{
		f->mem_offs = bump_file(f->mem_offs);
		bumpfile = f;
	}

	write_file(f->mem_offs, f->pos, src, size);
	f->pos += size;
	get_fileinfo(f->mem_offs, &length, NULL);
	f->length = length;
	return size;
}

void rdev_close(int fd)
{
	RFile *f = open_files[fd];
	if(f == bumpfile)
		bumpfile = NULL;
	free(f);
	open_files[fd] = NULL;

	dump_files();
}

int rdev_lseek(int fd, int offset, int origin)
{
	RFile *f = open_files[fd];


	switch(origin)
	{
		case PG_SEEK_SET:
			//printf("Seek to %d\n", offset);
			f->pos = offset;
			break;
		case PG_SEEK_CUR:
			//printf("Seek %d\n", offset);
			f->pos += offset;
			break;
		case PG_SEEK_END:
			//printf("Seek %d from end\n", offset);
			f->pos = f->length - offset;
	}

	if(f->pos < 0)
		f->pos = 0;
	if(f->pos > f->length)
		f->pos = f->length;

	return f->pos;
}

int rdev_tell(int fd)
{
	RFile *f = open_files[fd];
	return f->pos;
}

int rdev_ioctl(int fd, int request, ...)
{
	return -1;
}

int rdev_remove(const char *name)
{
	int offs = find_file(name);
	if(offs)
		remove_file(offs);
}

static int current_offs = 0;

int rdev_dopen(const char *name)
{
	current_offs = first_file();
}

int rdev_dread(int fd, void *d, int len)
{
	char *dest = (char*)d;

	if(!current_offs)
		return -1;

	get_fileinfo(current_offs, NULL, dest);
	current_offs = next_file(current_offs);

	return strlen(dest);
}

void rdev_dclose(int fd)
{
}

int rdev_stat(const char *name, struct pogo_stat *buffer)
{
	int offs = find_file(name);
	if(offs)
	{
		int length;
		get_fileinfo(offs, &length, NULL);
		buffer->pst_size = length;
		buffer->pst_mode = 0;
		return 0;
	}
	return -1;
}

void rdev_init(void *mem, int size)
{
	memset(&rdev, 0, sizeof(PogoDevice));
	rdev.open = rdev_open;
	rdev.read = rdev_read;
	rdev.write = rdev_write;
	rdev.close = rdev_close;
	rdev.lseek = rdev_lseek;
	rdev.stat = rdev_stat;
	rdev.dopen = rdev_dopen;
	rdev.dread = rdev_dread;
	rdev.dclose = rdev_dclose;
	rdev.remove = rdev_remove;

//	rinit(malloc(655360), 655360);
	rinit(mem, size, 0);

//	dump_files();

	pogo_device_register(&rdev, "/ram", NULL, -1);
}

#if RDEV_TEST

static uchar data[65536];

void make_file(char *name, int size)
{
	int i,fd;
	int c = 123;
	int total = 0;

	memset(data, 0, 65536);

	for(i=2; i<size; i++)
	{
		data[i] = (c & 0xff);
		total += c;
		c = (c+11) + ((c-3)<<3);
	}

	data[0] = total & 0xff;
	data[1] = total>>8;


	fd = rdev_open(name, PGO_WRONLY);
	if(fd >= 0)
	{
		if(size > 150)
		{
			rdev_write(fd, data, 143);
			rdev_write(fd, &data[143], size-143);
		}
		else
			rdev_write(fd, data, size);
		rdev_close(fd);
	}
}


int verify_file(char *name, int esize)
{
	int i,fd;
	int c = 123;
	int total = 0;

	memset(data, 0x55, 65536);

	fd = rdev_open(name, PGO_RDONLY);
	if(fd >= 0)
	{
		int size = rdev_read(fd, data, 65536);
		if(esize >=0 && size != esize)
			return 0;

		rdev_close(fd);

		for(i=2; i<size; i++)
		{
			if(data[i] != (c & 0xff))
				return 0;
			total += c;
			c = (c+11) + ((c-3)<<3);
		}
		return 1;
	}

	return 0;
}

char *make_name(int n)
{
	static char tmp[80];
	sprintf(tmp, "file%d", n);
	return tmp;
}

int grow_shrink_test(int c, int size, int v)
{
	int i;
	int t = 0;
	int sizes[256];

	dump_files();

	for(i=0; i<c; i++)
	{
		sizes[i] = size - (v/2) + (rand() % v);
		make_file(make_name(i), sizes[i]);
		t += verify_file(make_name(i), sizes[i]);
	}

	dump_files();

	for(i=0; i<c; i++)
	{
		sizes[i] = size - (rand() % v);
		make_file(make_name(i), sizes[i]);
		t += verify_file(make_name(i), sizes[i]);
	}

	for(i=0; i<c; i++)
	{
		make_file(make_name(i), size + (rand() % v));
		t += verify_file(make_name(i), -1);
	}

	for(i=0; i<c; i++)
		remove_file(make_name(i));

	return (t == c*3);
}

void remove_add_test()
{
}

void test_fs()
{
	int c,i;

	return;
	srand(time(NULL));

	int ok = grow_shrink_test(75, 120, 100);

}

#endif
