
#include <pogo/device.h>
#include <pogo/editbuf.h>
#include <pogo/console.h>
#include <pogo/readline.h>
//#include <pogo/ds_defs.h>


#define MAX_HIST 32

char *h_strings[MAX_HIST];
int history_pos = -1;

void history_init()
{
	int i;
	history_pos = 0;
	for(i=0; i<MAX_HIST; i++)
		h_strings[i] = POGO_NULL;
}

char *hstrdup(char *s)
{
	char *p = malloc(strlen(s)+1);
	strcpy(p, s);
	return p;
}




void history_shift()
{
	while(history_pos >= MAX_HIST)
	{
		int i;
		free(h_strings[0]);
		for(i=0; i<MAX_HIST-1; i++)
			h_strings[i] = h_strings[i+1];
		h_strings[i] = POGO_NULL;
		history_pos--;
	}
}



void history_set(char *line)
{
	if(history_pos == -1)
		history_init();
	
	history_shift();
	
	if(h_strings[history_pos])
		free(h_strings[history_pos]);
	h_strings[history_pos] = hstrdup(line);
}

void history_push(char *line)
{
	while(history_pos < (MAX_HIST-1) && h_strings[history_pos])
		history_pos++;
	
	history_shift();
	
	if(history_pos > 0 && strcmp(h_strings[history_pos-1], line) == 0)
		return;
		
	history_set(line);
	history_pos++;
}


char *history_up()
{
	if(history_pos == -1)
		history_init();

	if(history_pos)
		return h_strings[--history_pos];
	else
		return POGO_NULL;
}

char *history_down()
{
	if(history_pos == -1)
		history_init();
	
	if(history_pos < (MAX_HIST-1) && h_strings[history_pos+1])
		return h_strings[++history_pos];
	else
	{
		return POGO_NULL;
	}
}




static char editbuf[1024];
static int oldlen;
static int oldx;
static int width;

void readline_start(int w)
{
	int size = pogo_ioctl(POGO_STDOUT, CC_GETPOS); //console_getcharcount();
	editbuf_init(editbuf, sizeof(editbuf));
	pogo_ioctl(POGO_STDOUT, CC_MARK, size, size);
	oldx = 0;
	oldlen = 0;
	width = w;
}

char *readline_end(int clear)
{
	pogo_ioctl(POGO_STDOUT, CC_MARK, -1, -1);
	if(clear)
	{
		editbuf_seek(0, PG_SEEK_END);
		pogo_ioctl(POGO_STDOUT, CC_ERASE, strlen(editbuf_get()));
	}

	history_push(editbuf_get());
	return editbuf_get();
}

static const char **dict_words = POGO_NULL;
static int dict_count = 0;
static int hitlist[80];
static int hits = 0;
static int hit_pos;
static int last_hit = 0;

void readline_set_dict(const char **words, int count)
{
	dict_words = words;
	dict_count = count;
}

void readline_complete()
{
	int l;
	if(hits)
	{
		last_hit++;
		if(last_hit >= hits)
			last_hit = 0;
		
		editbuf_del(-1, editbuf_tell() - hit_pos);
		editbuf_write(dict_words[hitlist[last_hit]], strlen(dict_words[hitlist[last_hit]]));
		editbuf_put(' ');

	}
	else
	{
		int i;
		char *name = editbuf_get();
		char *p = &name[strlen(name)-1];
		while(p > name)
		{
			if(*p == ' ')
			{
				p++;
				break;
			}
			p--;
		}
		
		name = p;

		l = strlen(name);
		hits = 0;
		for(i=0; i<dict_count; i++)
		{
			if(strncmp(name, dict_words[i], l) == 0)
			{
				hitlist[hits++] = i;
			}
		}
		
		if(hits > 0)
		{
			editbuf_del(-1, l);
			hit_pos = editbuf_tell();
			editbuf_write(dict_words[*hitlist], strlen(dict_words[*hitlist]));
			last_hit = 0;
			editbuf_put(' ');
			if(hits == 1)
			{
				hits = 0;
			}
		}
	}
}
		
void readline_refresh()
{
	int sl,  size;
	char *tx;
	int x = editbuf_seek(0, PG_SEEK_CUR);
	pogo_ioctl(POGO_STDOUT, CC_ERASE, oldlen);
	tx = editbuf_get();
	sl = strlen(tx);
	pogo_write(POGO_STDOUT, tx, sl);
	size = pogo_ioctl(POGO_STDOUT, CC_GETPOS); //console_getcharcount();
	pogo_ioctl(POGO_STDOUT, CC_MARK, size-sl+x, size-sl+x);
	
	//oldlen = strlen(editbuf_get());	
	//sl = strlen(editbuf_get());
	//size = pogo_ioctl(POGO_STDOUT, CC_GETPOS); //console_getcharcount();
	//pogo_ioctl(1, CC_MARK, size-sl+x, size-sl+x);

}

int readline_putchar(int c)
{
	int x;

	if(c)
	{
		switch(c)
		{
		case RL_KEY_RIGHT:
			editbuf_seek(1, PG_SEEK_CUR);
			break;
		case RL_KEY_LEFT:
			editbuf_seek(-1, PG_SEEK_CUR);
			break;
		case RL_KEY_UP:
			//history_set(editbuf_get());
			editbuf_set(history_up());
			editbuf_seek(0, PG_SEEK_END);
			hits = 0;
			break;
		case RL_KEY_DOWN:
			//history_set(editbuf_get());
			editbuf_set(history_down());
			editbuf_seek(0, PG_SEEK_END);
			hits = 0;
			break;
		case RL_KEY_HOME:
			editbuf_seek(0, PG_SEEK_SET);
			break;
		case RL_KEY_END:
			editbuf_seek(0, PG_SEEK_END);
			break;
		case RL_KEY_DEL:
			editbuf_del(-1, 1);
			hits = 0;
			break;
		case RL_KEY_KILL:
		{
			char *s = editbuf_get();
			char *e = &s[strlen(s)-1];
			while(e > s && *e == ' ') e--;
			while(e > s && *e != ' ') e--;
			if(*e == ' ') e++;
			*e = 0;
			editbuf_set(s);
			//if(e-s > 0)
//				editbuf_del(-1, e-s);
			hits = 0;
			break;
		}
		case RL_KEY_COMPLETE:
			readline_complete();
			break;				
		//case SDLK_ENTER:
		//	putchar(10);
		//	break;
		default:
			hits = 0;
			editbuf_put(c);
			break;
		}
	}

	x = editbuf_seek(0, PG_SEEK_CUR);

	if(editbuf_changed())
	{
		char *tx;
		int sl,size;
		//int oldf = 0;//console_ioctl(1, CC_GETFONT);
		//console_erase(oldlen);

		pogo_ioctl(POGO_STDOUT, CC_ERASE, oldlen);
		tx = editbuf_get();
		sl = strlen(tx);
		//printf(tx);
		pogo_write(POGO_STDOUT, tx, sl);
		size = pogo_ioctl(POGO_STDOUT, CC_GETPOS); //console_getcharcount();

		//console_mark(size-sl+x, size-sl+x);
		pogo_ioctl(POGO_STDOUT, CC_MARK, size-sl+x, size-sl+x);
		oldlen = strlen(tx);//editbuf_get());
	}
	else
	if(x != oldx)
	{
		int sl = strlen(editbuf_get());
		int size = pogo_ioctl(POGO_STDOUT, CC_GETPOS); //console_getcharcount();
		//console_mark(size-sl+x, size-sl+x);
		pogo_ioctl(1, CC_MARK, size-sl+x, size-sl+x);
		oldx = x;
	}
	else
		return 0;
	return 1;

}

//static uint16 oldkeys = 0;

/*
int readline_update()
{
	uint16 keys;
	int c = EOF;

	keys = ~(*KEYS);

	if(keys != oldkeys)
	{
		oldkeys = keys;

		if(keys & 0x10)
			c = RL_KEY_RIGHT;
		else
		if(keys & 0x20)
			c = RL_KEY_LEFT;
		else
		if(keys & 0x80)
			c = RL_KEY_DOWN;
		else
		if(keys & 0x40)
			c = RL_KEY_UP;
		else
		if(keys & 0x1)
			c = RL_KEY_ENTER;
		else
		if(keys & 0x2)
			c = RL_KEY_DEL;
		else
		if(keys & 0x8)
			c = 'c';
		else
		if(keys & 0x4)
			c = 'd';
		else
		if(keys & 0x100)
			c = 'e';
		else
		if(keys & 0x200)
			c = 'f';
	}

	if(c == RL_KEY_ENTER)
		return 1;

	if(c == EOF)
		c = getchar();

	if(c != EOF)
		readline_putchar(c);

	return 0;
}
*/

void readline_set(char *str)
{
	editbuf_set(str);
	readline_putchar(0);
}
