#include <nds.h>

#include <pogo/device.h>

#undef NDS
#define NDS
#include "fat.h"
#undef PACKED

#include <sys/dir.h>
#include <stdio.h>
#include <string.h>

PogoDevice fatdev;

static FILE *open_files[32];
static int init = 0;

char tmpname[256];

int fatdev_open(const char *name, int flags)
{
	FILE *f;

	if(!init)
	{
		memset(open_files, 0, sizeof(FILE*) * 32);
		init = 1;
	}

	//mod_volume(0);

	chdir("/");

	while(*name == '/') name++;
	sprintf(tmpname, "/%s", name);

	char *fg;

	if(flags == PGO_RDONLY)
		fg = "r";
	else
	if(flags == PGO_RDWR)
		fg = "r+";
	else
	if(flags == PGO_RDWR|PGO_CREAT)
		fg = "w+";
	else
	if(flags & PGO_WRONLY)
		fg = "w";


	f = fopen(tmpname, fg);

	if(f)
	{
		int i = 0;
		while(open_files[i]) i++;
		open_files[i] = f;
		return i;
	}
	//mod_volume(127);
	return -1;
}

int fatdev_read(int fd, void *dest, int size)
{
	FILE *f = open_files[fd];
	return fread(dest, 1, size, f);
}

int fatdev_write(int fd, const void *dest, int size)
{
	FILE *f = open_files[fd];
	return fwrite(dest, 1, size, f);
}

void fatdev_close(int fd)
{
	FILE *f = open_files[fd];
	fclose(f);
	open_files[fd] = NULL;
	//mod_volume(127);
}

int fatdev_lseek(int fd, int offset, int origin)
{
	FILE *f = open_files[fd];

	fseek(f, offset, origin);
	return ftell(f);
}

int fatdev_tell(int fd)
{
	FILE *f = open_files[fd];
	return ftell(f);
}

int fatdev_ioctl(int fd, int request, ...)
{
	return -1;
}

int fatdev_remove(const char *name)
{
	chdir("/");
	while(*name == '/') name++;
	sprintf(tmpname, "/%s", name);
	return remove(tmpname);
}

int fatdev_mkdir(const char *name)
{
	chdir("/");
	while(*name == '/') name++;
	sprintf(tmpname, "/%s", name);
	return mkdir(tmpname, S_IRWXU);
}

static DIR_ITER *opendirs[16]; // inspired by native functions
static int dcount = 0;

int fatdev_dopen(const char *name)
{
// inspired by native functions
	DIR_ITER *d = diropen(name);
	if(d)
	{
		opendirs[dcount++] = d;
		return dcount-1;
	}
	return -1;
}

int fatdev_dread(int fd, void *d, int len)
{
// inspired by native functions
  char dest[256];
  struct stat filestat;
	DIR_ITER *dir = opendirs[fd];

	if (dirnext(dir,dest,&filestat) == -1) 
	{
	  strcpy((char *)d, "(null)");
		return 0;
	//return -1; ?
	} 
	else
	{
		strncpy((char *)d, dest, len);
		return strlen(dest);
	}
}

void fatdev_dclose(int fd)
{
// inspired by native functions
	dirclose(opendirs[fd]);
	opendirs[fd] = NULL;

	while(dcount && !opendirs[dcount-1])
		dcount--;
}

int fatdev_stat(const char *name, struct pogo_stat *buffer)
{
// inspired by native functions
	struct stat s;

	int rc = stat(name, &s);
	if(rc == 0)
	{
		buffer->pst_size = s.st_size;
		buffer->pst_atime = (uint32)s.st_atime;
		if(s.st_mode & S_IFDIR)
			buffer->pst_mode = PGS_IFDIR;
		else
			buffer->pst_mode = 0;
	}

	return rc;
}

int fatdev_init()
{
	memset(&fatdev, 0, sizeof(PogoDevice));
	fatdev.open = fatdev_open;
	fatdev.read = fatdev_read;
	fatdev.write = fatdev_write;
	fatdev.close = fatdev_close;
	fatdev.lseek = fatdev_lseek;
	fatdev.stat = fatdev_stat;
	fatdev.dopen = fatdev_dopen;
	fatdev.dread = fatdev_dread;
	fatdev.dclose = fatdev_dclose;
	fatdev.remove = fatdev_remove;
	fatdev.mkdir = fatdev_mkdir;

	int rc = fatInitDefault();

	if(rc)
		pogo_device_register(&fatdev, "/fat", NULL, -1);

	return rc;
}
