/*
 * Simple SDL Keyboard Device (sdlkey.c)
 * Part of libpogo, a c-library replacent for GBA
 * Programmed by Jonas Minnberg (Sasq)
 *
 * DESCRIPTION
 * Stdin device for SDL
 *
 **/

#include <SDL.h>
#include "pogo/device.h"
#include "pogo/screen.h"

static int new_width;
static int new_height;
static int resize;

static int modifiers = 0;

extern volatile unsigned char *inptr;
extern volatile uint16 *clickptr;


void display_dump(char *fname);
static int shotn = 1;

extern int strokes;
void end_stroke();
void do_stroke(int x, int y);
extern int stroke_ended;

extern int sa_x0;
extern int sa_x1;
extern int sa_y0;
extern int sa_y1;
extern void *sa_ptr;
extern int sa_flags;


static int count = 0;
static int last_tick = 0;

static int lastx = -1;
static int lasty;

//int touchx;
//int touchy;
//int buttons;

void convert_xy(int *x, int *y);


int key_get_pointer(int *b, int *x, int *y)
{
	int bt, w, h, r, xx, yy;
	
	
	SDL_PumpEvents();

	bt = SDL_GetMouseState(x, y);
	convert_xy(x, y);

	if(b)
		*b = (bt & SDL_BUTTON(1));
		
	return 0;
				
}

static uchar unicodes[256];

int sdlevents_get()
{
	int c,ch;
	SDL_Event event;
	SDL_PumpEvents();

	if(SDL_PollEvent(&event))
	{
		
		int tick = SDL_GetTicks();
		switch(event.type)
		{
		case SDL_MOUSEMOTION:
			if(SDL_BUTTON(event.motion.state) == 1)// && (last_tick + 10 <= tick))
			{
				count++;
				int x = event.motion.x;
				int y = event.motion.y;
				
				convert_xy(&x, &y);
				
				last_tick = tick;

				if(x >= sa_x0 && x < sa_x1 && y >= sa_y0 && y < sa_y1 && !stroke_ended && count >= 2)
				{					
					
					if(sa_ptr && !strokes)
						screen_save();
					
					do_stroke(x,  y);
					
					//display_set(0);
					//if(lastx >= 0)
						//draw_line(lastx, lasty, x, y);
					lastx = x;
					lasty = y;
					display_flush();
					
					//printf("%d %d\n",  event.motion.x,  event.motion.y);
					//strokes++;
				}
			}

			break;
		case SDL_MOUSEBUTTONUP:
			if(event.button.button == SDL_BUTTON_LEFT)
			{
				count = 0;
				//printf("strokes %d\n", strokes);
				if(strokes)
				{
					stroke_ended = 1;
					if(sa_ptr)
					{
						screen_load();
						display_flush();
					}
				}
				lastx = -1;
			}
			break;
		case SDL_MOUSEBUTTONDOWN:
			if(event.button.button == SDL_BUTTON_LEFT)
			{
				int i;
				int x = event.button.x;
				int y = event.button.y;
				
				convert_xy(&x, &y);

				*clickptr++ = x;
				*clickptr++ = y;
			}
			break;
		case SDL_QUIT:
			exit(0);
			return -1;

		case SDL_VIDEORESIZE:
			new_width = event.resize.w & 0xFFFFFFFC;
			new_height = event.resize.h;
			resize = 1;
			return 0;

		case SDL_KEYUP:
		case SDL_KEYDOWN:
			c = 0;
			ch = event.key.keysym.sym;
			switch(ch)
			{
				case SDLK_F12:
					display_dump("screenshot.bmp");
					break;
				//case SDLK_LSHIFT:
				//	c = RAWKEY_L;
				//	break;
				case SDLK_PAGEDOWN:
					c = POGOKEY_PAGEDOWN;
					break;
				case SDLK_END:
					c = POGOKEY_END;
					break;
				case SDLK_HOME:
					c = POGOKEY_HOME;
					break;
				case SDLK_DELETE:
					c = POGOKEY_DEL;
					break;
				case SDLK_TAB:
					c = POGOKEY_TAB;
					break;
				case SDLK_UP:
					c = POGOKEY_UP;
					break;
				case SDLK_DOWN:
					c = POGOKEY_DOWN;
					break;
				case SDLK_LEFT:
					c = POGOKEY_LEFT;
					break;
				case SDLK_RIGHT:
					c = POGOKEY_RIGHT;
					break;
				case SDLK_INSERT:
					c = POGOKEY_INSERT;
					break;
				case SDLK_PAGEUP:
					c = POGOKEY_PAGEUP;
					break;
				case SDLK_RETURN:
					c = POGOKEY_ENTER;
					break;
				case SDLK_BACKSPACE:
					c = POGOKEY_BACKSPACE;
					break;
				case SDLK_ESCAPE:
					c = POGOKEY_ESCAPE;
					break;
				case SDLK_KP1:
					c = 'b'; break;
				case SDLK_KP2:
					c = 'j'; break;
				case SDLK_KP3:
					c = 'n'; break;
				case SDLK_KP4:
					c = 'h'; break;
				case SDLK_KP5:
					c = '.'; break;
				case SDLK_KP6:
					c = 'l'; break;
				case SDLK_KP7:
					c = 'y'; break;
				case SDLK_KP8:
					c = 'k'; break;
				case SDLK_KP9:
					c = 'u'; break;
				default:					
					if(ch >= 0x20 && ch <=0x7f)
					{
						if(event.type == SDL_KEYDOWN)
						{
							c = event.key.keysym.unicode;
							if(c > 0x7f)
								c = 0;
							unicodes[ch] = c;
						}
						else
						{
							c = unicodes[ch];
						}
					}
			}
			/*			
			if(event.key.keysym.mod & KMOD_SHIFT)
			{
				int i;
				if(c >= 'a' && c <= 'z')
					c -= 0x20;
				else
				{
					static char conv[] = "0)1!2@3#4$5%6^7&8*9(-_=+[{]};:'\"\\|,<.>/?`~";

					for(i=0; i<sizeof(conv); i+=2)
					{
						if(c == conv[i])
						{
							c = conv[i+1];
							break;
						}
					}
				}

			}
			*/

			if(c)
			{
				if(event.type == SDL_KEYUP)
					c |= 0x80;
				*inptr++ = c;
			}
			//if(event.key.keysym.sym == SDLK_ESCAPE)
			//	return 1;
			//break;				
		}
	}
	//else
	//if(resize)
	//{
	//	screen_set_target(NULL, new_width, new_height, -1);
	//	resize = 0;
	//}

	return 0;
}

int sdlkey_get_modifiers()
{
	int m = 0;

	int mod = SDL_GetModState();
	if(mod & (KMOD_LCTRL|KMOD_RCTRL))
		m |= 2;
	if(mod & (KMOD_LSHIFT|KMOD_RSHIFT))
		m |= 1;

	return m;
}

int sdlkey_peek()
{
	SDL_Event event[10];
	SDL_PumpEvents();
	int rc = SDL_PeepEvents(event, 10, SDL_PEEKEVENT, SDL_EVENTMASK (SDL_KEYDOWN));
	return (rc > 0);
}

void sdlkey_init()
{
	SDL_EnableUNICODE(1);
	//SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY, SDL_DEFAULT_REPEAT_INTERVAL);
	SDL_EnableKeyRepeat(0, 0);
}
