#include <pogo/platform.h>

#ifdef POGO_PLATFORM_NDS
#include <nds.h>
#include <dswifi9.h>
#endif

#ifdef USE_WINSOCK
#include <winsock2.h>
#else
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#endif


#include <pogo/pogo.h>
#include <pogo/font.h>
#include <pogo/console.h>
#include <pogo/keyboard.h>
#include <pogo/io.h>
#include <pogo/readline.h>

#include <sys/types.h>
#include <sys/stat.h>

#include <stdio.h>
#include <stdlib.h>

// SDL.h must be included where main() is defined on some platforms (like MACOSX)
#ifdef USE_SDL
#include <SDL.h>
#endif

void native_init();
void gbfs_init();

// 10 buttons wide keyboard
char *keyb_def[] = {
	" x3 12 22 32 f1[{2]}2-_2=+2'\"2",
	"1!22@23#24$25%26^27&28*29(20)2",
	"qQ2wW2eE2rR2tT2yY2uU2iI2oO2pP2",
	"aA2sS2dD2fF2gG2hH2jJ2kK2lL2;:2",
	"zZ2xX2cC2vV2bB2nN2mM2,<2.>2/?2",
	" s2 C2 t3  7 b3 e3",
	POGO_NULL
};

int screen_width, screen_height, screen_stride, screen_rotation;
Font *textfont, *keyfont;

int titlebar = -1;
int mainwin = -1;

void setup_windows()
{
	Font *fonts[4] = { POGO_NULL, POGO_NULL, POGO_NULL, POGO_NULL };
	int h = screen_height - 24;
	fonts[0] = fonts[1] = textfont;	
	
	if(titlebar >= 0)
	{
		pogo_ioctl(POGO_STDOUT, CC_SETWIN, titlebar);
		pogo_ioctl(POGO_STDOUT, CC_MOVEWIN, 0, 0, screen_width, 24);
		pogo_ioctl(POGO_STDOUT, CC_SETWIN, mainwin);
		pogo_ioctl(POGO_STDOUT, CC_MOVEWIN, 0, 24, screen_width, h);	
	}
	else
	{
		titlebar = pogo_ioctl(POGO_STDOUT, CC_OPENWIN, 0, 0, screen_width, 24, fonts, CONWIN_STREAM);
		pogo_ioctl(POGO_STDOUT, CC_SETCOLOR, CCOL_WHITE, CCOL_BLUE);
		pogo_ioctl(POGO_STDOUT, CC_CLEAR);
		pogo_printf("Hello people");
		mainwin = pogo_ioctl(POGO_STDOUT, CC_OPENWIN, 0, 24, screen_width, h, fonts, CONWIN_STREAM);
		pogo_ioctl(POGO_STDOUT, CC_SETCOLOR, CCOL_BLACK, CCOL_WHITE);
		pogo_ioctl(POGO_STDOUT, CC_CLEAR);

	}
}

enum {
#ifdef POGO_PLATFORM_NDS
	IFCONFIG,
	WFCCONNECT,
	LISTAP,
	STATUS,
	SCAN,
	CONNECT,
	DISCONNECT,
#endif
	GETURL,
	HELP,
	MAXCMD
};

const char *commands[] = {
#ifdef POGO_PLATFORM_NDS
	 "ifconfig", "wfcconnect", "listap", "status", "scan", "connect", "disconnect", 
#endif
	 "geturl", "help"

};


const char *statusnames[] = {
	"DISCONNECTED",
	"SEARCHING",
	"AUTHENTICATING",
	"ASSOCIATING",
	"ACQUIRING DHCP",
	"ASSOCIATED",
	"CANNOT CONNECT",
};

void print_ip(unsigned int ip)
{
	pogo_printf("%d.%d.%d.%d", ip&0xff, (ip>>8)&0xff, (ip>>16)&0xff, (ip>>24)&0xff);
}





void get_url(char *url)
{
	char server[80];
	char request_text[256];
	char *p = url;
	
	while(*p && *p != '/') p++;
	if(!*p)
		return;

	strncpy(server, url, p-url);
	server[p-url] = 0;
	sprintf(request_text,
			"GET %s HTTP/1.1\r\n"
			"Host: %s\r\n"
			"User-Agent: Nintendo DS\r\n\r\n",
			p, server);
	
	
	pogo_printf("'%s' '%s'\n", server, p); 
	display_flush();
		
    // Find the IP address of the server, with gethostbyname
	struct hostent * myhost = gethostbyname( server );
	if(!myhost)
	{
		pogo_printf("Could not resolve '%s' !\n", server);
		return;
	}
	pogo_printf("Found IP Address!\n");
	display_flush();
 
    // Create a TCP socket
	int my_socket;
	my_socket = socket( AF_INET, SOCK_STREAM, 0 );
	pogo_printf("Created Socket!\n");
	display_flush();

    // Tell the socket to connect to the IP address we found, on port 80 (HTTP)
	struct sockaddr_in sain;
	sain.sin_family = AF_INET;
	sain.sin_port = htons(80);
	sain.sin_addr.s_addr= *( (unsigned long *)(myhost->h_addr_list[0]) );
	connect( my_socket,(struct sockaddr *)&sain, sizeof(sain) );
	pogo_printf("Connected to server!\n");
	display_flush();

    // send our request
	send( my_socket, request_text, strlen(request_text), 0 );
	pogo_printf("Sent our request!\n");
	display_flush();

    // Print incoming data
	pogo_printf("Printing incoming data:\n");
	display_flush();

	int recvd_len;
	char incoming_buffer[256];
	
	int c = 0;

	while( ( recvd_len = recv( my_socket, incoming_buffer, 255, 0 ) ) != 0 ) { // if recv returns 0, the socket has been closed.
		if(recvd_len>0) { // data was received!
			incoming_buffer[recvd_len] = 0; // null-terminate
			pogo_printf(incoming_buffer);
			display_flush();
		}
		
		c = pogo_getchar();
		if(c == POGOKEY_ESCAPE)
			break;
	}

	if(c != POGOKEY_ESCAPE)
		pogo_printf("Other side closed connection!");
	display_flush();

	shutdown(my_socket,0); // good practice to shutdown the socket.

	close(my_socket); // remove the socket.

}








void parse_line(const char *line)
{
	int i,rc;
	int cmd = -1;
	char tmp[128];
	char *arg = NULL;
	char *p = tmp;
	strcpy(tmp, line);
	while(*p && *p != ' ') p++;
	if(*p == ' ')
	{
		*p = 0;
		arg = p+1;
	}
		
	
	for(i=0; i<MAXCMD; i++)
		if(strcmp(tmp, commands[i]) == 0)
		{
			cmd = i;
			break;
		}
		
	pogo_puts("");
	
	switch(cmd)		
	{
#ifdef POGO_PLATFORM_NDS
		case IFCONFIG:
		{
			unsigned long gw, mask, dns1, dns2;
			u32 ip = (u32)Wifi_GetIP();
			Wifi_GetIPInfo(&gw, &mask, &dns1, &dns2);
			pogo_printf("\nIP: ");
			print_ip(ip);
			pogo_printf("\nMASK: ");
			print_ip(mask);
			pogo_printf("\nGW: ");
			print_ip(gw);
			pogo_printf("\nDNS1: ");
			print_ip(dns1);
			pogo_printf("\nDNS2: ");
			print_ip(dns2);
			break;
		}
		case WFCCONNECT:
			Wifi_AutoConnect(); // request connect	
			pogo_printf("Trying to connect via WFC (%s)\n", statusnames[Wifi_AssocStatus()]);
			break;
		case SCAN:
			pogo_printf("Entering scanmode\n");
			Wifi_ScanMode();
			break;			
		case STATUS:
			rc = Wifi_AssocStatus();
			pogo_printf("%s\n", statusnames[rc]);
			break;
		case LISTAP:
		{	Wifi_AccessPoint as;
			int na = Wifi_GetNumAP();
			
			pogo_puts("");
			
			for(i=0; i<na; i++)
			{
				Wifi_GetAPData(i, &as);
				pogo_printf("[%d] %s (%d)\n", i, as.ssid, as.rssi);
			}
			break;
		}
		case DISCONNECT:
			pogo_printf("Disconnecting\n");
			Wifi_DisconnectAP();
			break;
		case CONNECT:
		{	Wifi_AccessPoint as;
			int na = Wifi_GetNumAP();
								
			for(i=0; i<na; i++)
			{
				Wifi_GetAPData(i, &as);
				
				if(strcmp(arg, as.ssid) == 0)
				{
					printf("Connecting to %s\n", as.ssid);
					Wifi_ConnectAP(&as, WEPMODE_NONE, 0, NULL);
				}
			}
			break;
		}
#endif
		case HELP:
			pogo_puts("");
			for(i=0; i<MAXCMD; i++)
				pogo_printf("%s\n", commands[i]);
			break;
		case GETURL:
			get_url(arg);
			break;
	}
	
}
		
		
int last_status = -123;

int main(int argc, char **argv)
{
	int cfd, mainw,c;
	uint16 *screen_pixels;
	int i;

#ifdef USE_WINSOCK
	WSADATA wsa;
	WORD ver = MAKEWORD(1, 1);
	WSAStartup(ver, &wsa);
#endif


	pogo_device_init();

#ifdef USE_SDL
	native_init();
#else
	if(!gbfsdev_init())
		fatdev_init();
#endif

	screen_width = 640;
	screen_height = 480;	
	
	// set size is ignored on platforms with fixed screen - ie anything besides PC
	display_set_params(screen_width, screen_height, 0);
	display_init();	
	
	screen_rotation = ROT_TOPLEFT;
	screen_set_rotation(screen_rotation);
			
	screen_get_target(NULL, &screen_width, &screen_height, POGO_NULL);	
				
	key_init();
	console_init();

	// The first file we open, makes sure they become STDIN and STDOUT respectivly
	pogo_open("/dev/key", 0);				
	pogo_open("/dev/console", 0);
	
	textfont = font_getdefault();
	
	// Try to load a font but fall back to the internal default font
	keyfont = font_load("Arial.font");
	if(!keyfont) 
		keyfont = font_getdefault();
	
	setup_windows();

	void *ptr;

#ifdef POGO_PLATFORM_PC
	// On PC we have a mouse and can set up a gesture area over the entire screen
	screen_get_target(&ptr, POGO_NULL, POGO_NULL, POGO_NULL);	
	pogo_ioctl(POGO_STDIN, KC_SETUP_GESTURE_TARGET, ptr, 0, 24, screen_width, screen_height-24, 0);
#endif
	
	pogo_puts("POGO WIFI TEST\n");
	pogo_puts("SELECT to flip upper/lower. X to rotate display\n'help' for commands.\n");
	
#ifdef POGO_PLATFORM_NDS
	// On NDS we set up a virtual keyboard on the lower screen
	ptr = 0x06200000;
	screen_set_target(0x06200000, 0, 0);
	screen_set(0, 0, screen_width, screen_height, UINT2PIXEL(0x8080E0));	
	int kh = pogo_ioctl(POGO_STDIN, KC_SETUP_KEYBOARD, 0, 14, screen_width, keyb_def, keyfont);

	pogo_puts("Trying to init network");
	netlib_init();
#endif

	display_flush();
	
	
	int draw = 0;
	int text_grab = 0;
	
	c = 0;
	pogo_getchar();
	
	readline_set_dict(commands, MAXCMD);
	
	
	pogo_putchar('>');
	readline_start(200);
	display_flush();
	
	while(1)//c != 27)
	{
		int b,x,y,ch = 0;
		
		//key_get_pointer(&b, &x, &y);
/*		pogo_ioctl(POGO_STDIN, KC_GET_POINTER, &b, &x, &y);
		if(b)
		{
			if(!text_grab && y < 10)
			{
				pogo_ioctl(POGO_STDIN, KC_SETUP_GESTURE_TARGET, NULL, -1, -1, -1, -1, 0);
				text_grab = 1;
				
			}
		}
		else
		{
			if(text_grab)
			{
				char temp[80];
				char *p = temp;
				if(console_findword(x, y, temp))
				{				
					while(*p)
						readline_putchar(*p++);
				}
				pogo_ioctl(POGO_STDIN, KC_SETUP_GESTURE_TARGET, ptr, 0, 24, screen_width, screen_height-24, 0);
				text_grab = 0;
				
				console_markword(-1, -1);
				
				display_flush();
			}

		}
		
		if(text_grab)
		{
			console_markword(x, y);
			display_flush();
		}
*/
		
#ifdef POGO_PLATFORM_NDS
		int rc = Wifi_AssocStatus();
		
		if(rc != last_status)
		{
			last_status = rc;
			pogo_ioctl(POGO_STDOUT, CC_SETWIN, titlebar);
			//pogo_ioctl(POGO_STDOUT, CC_GOTOXY, 0, 0);
			pogo_ioctl(POGO_STDOUT, CC_ERASE, 100);
			pogo_printf("STATUS: %s", statusnames[rc]);
			pogo_ioctl(POGO_STDOUT, CC_SETWIN, mainwin);

		}
#endif
		
		c = pogo_getchar();
		
		if(c > 0 && c < 0x80)
		{
			switch(c)
			{
				case RAWKEY_Y:
				case POGOKEY_TAB:
					ch = RL_KEY_COMPLETE;
					break;
				case RAWKEY_X:
				case POGOKEY_PAGEUP:
					screen_rotation++;
					if(screen_rotation == ROT_BOTTOMLEFT + 1)
						screen_rotation = ROT_TOPLEFT;
					screen_set_rotation(screen_rotation);
					screen_get_target(NULL, &screen_width, &screen_height, POGO_NULL);	
					setup_windows();
#ifdef POGO_PLATFORM_PC
					pogo_ioctl(POGO_STDIN, KC_SETUP_GESTURE_TARGET, ptr, 0, 24, screen_width, screen_height-24, 0);
#endif
					readline_refresh();
					display_flush();
					
#ifdef POGO_PLATFORM_NDS
					if(draw)					
						pogo_ioctl(POGO_STDIN, KC_SETUP_GESTURE_TARGET, 0x06000000, 0, 0, screen_width, screen_height, 0);

					ptr = 0x06200000;
					screen_set_target(0x06200000, 0, 0);
					screen_set(0, 0, screen_width, screen_height, UINT2PIXEL(0x8080E0));	
					int kh = pogo_ioctl(POGO_STDIN, KC_SETUP_KEYBOARD, 0, 14, screen_width, keyb_def, keyfont);
#endif
					
					break;
				case 8:
				case RAWKEY_B:
					ch = RL_KEY_DEL;
					break;
				case RAWKEY_LEFT:
					ch = RL_KEY_LEFT;
					break;
				case RAWKEY_RIGHT:
					ch = RL_KEY_RIGHT;
					break;
				case RAWKEY_UP:
					ch = RL_KEY_UP;
					break;
				case RAWKEY_DOWN:
					ch = RL_KEY_DOWN;
					break;
				case 10:
				case RAWKEY_START:
					parse_line(readline_end(0));
					pogo_putchar(10);
					pogo_putchar('>');
					readline_start(200);
					display_flush();
					ch = 0;
					break;

#ifdef POGO_PLATFORM_NDS
				case RAWKEY_SELECT:
					if(!draw)
					{
						lcdMainOnBottom();
						pogo_ioctl(POGO_STDIN, KC_SETUP_GESTURE_TARGET, 0x06000000, 0, 0, screen_width, screen_height, 0);
						pogo_ioctl(POGO_STDIN, KC_ACTIVATE_KEYBOARD, 0);
						draw =1 ;
					}
					else
					{
						lcdMainOnTop();
						//key_set_stroke_target(NULL, -1, -1, -1, -1, 0);
						pogo_ioctl(POGO_STDIN, KC_SETUP_GESTURE_TARGET, NULL, -1, -1, -1, -1, 0);
						pogo_ioctl(POGO_STDIN, KC_ACTIVATE_KEYBOARD, 1);
						draw = 0;
					}
#endif
				
				default:
					if(c >= 0x20)
						ch = c;
			}
			if(ch && readline_putchar(ch))
				display_flush();
		}
	}
	
	return 0;
}
